using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Security.AccessControl;
using System.Security.Principal;
using System.Threading.Tasks;

using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Acl;

using CalDAVServer.NtfsStorage.Acl;


namespace CalDAVServer.NtfsStorage
{
    /// <summary>
    /// Base class for WebDAV items (folders, files, etc).
    /// </summary>
    public abstract class DavHierarchyItem : Discovery , IHierarchyItemAsync, ICurrentUserPrincipalAsync
    {
        /// <summary>
        /// Instance of <see cref="FileSystemInfo"/> corresponding to a file represented by this object.
        /// </summary>
        internal FileSystemInfo fileSystemInfo;

        /// <summary>
        /// Instance of <see cref="DavContext"/>.
        /// </summary>
        protected DavContext context;

        /// <summary>
        /// Encoded WebDAV relative path to the item.
        /// </summary>
        private readonly string path;

        /// <summary>
        /// User defined property values.
        /// </summary>
        private List<PropertyValue> propertyValues;

        /// <summary>
        /// Initializes a new instance of the DavHierarchyItem class.
        /// </summary>
        /// <param name="fileSystemInfo">Corresponding <see cref="FileSystemInfo"/> to this file/folder.</param>
        /// <param name="context">Instance of <see cref="DavContext"/>.</param>
        /// <param name="path">Relative to WebDAV root folder path.</param>
        protected DavHierarchyItem(FileSystemInfo fileSystemInfo, DavContext context, string path): base(context)
        {
            this.fileSystemInfo = fileSystemInfo;
            this.context = context;
            this.path = path;
        }

        /// <summary>
        /// Gets name of the item.
        /// </summary>
        public string Name
        {
            get { return fileSystemInfo.Name; }
        }

        /// <summary>
        /// Gets date when the item was created in UTC.
        /// </summary>
        public DateTime Created
        {
            get { return fileSystemInfo.CreationTimeUtc; }
        }

        /// <summary>
        /// Gets date when the item was last modified in UTC.
        /// </summary>
        public DateTime Modified
        {
            get { return fileSystemInfo.LastWriteTimeUtc; }
        }

        /// <summary>
        /// Gets path of the item where each part between slashes is encoded.
        /// </summary>
        public virtual string Path
        {
            get { return path; }
        }

        /// <summary>
        /// Returns parent folder of the item.
        /// </summary>
        /// <returns>Parent folder.</returns>
        public async Task<DavFolder> GetParentAsync()
        {
            //Calculate local file system path of parent item.
            string parentLocalPath = fileSystemInfo.FullName.TrimEnd('\\');
            int index = parentLocalPath.LastIndexOf('\\');
            if (index <= context.RepositoryPath.Length - 1)
            {
                return null;
            }

            var directory = new DirectoryInfo(parentLocalPath.Remove(index));

            //Calculate WebDAV root relative parent path.
            string parentPath = Path.TrimEnd('/');
            int ind = parentPath.LastIndexOf('/');
            if (ind >= 0)
            {
                parentPath = parentPath.Substring(0, ind);
            }

            return (DavFolder) await context.GetHierarchyItemAsync(parentPath);
        }

        /// <summary>
        /// Creates a copy of this item with a new name in the destination folder.
        /// </summary>
        /// <param name="destFolder">Destination folder.</param>
        /// <param name="destName">Name of the destination item.</param>
        /// <param name="deep">Indicates whether to copy entire subtree.</param>
        /// <param name="multistatus">If some items fail to copy but operation in whole shall be continued, add
        /// information about the error into <paramref name="multistatus"/> using 
        /// <see cref="MultistatusException.AddInnerException(string,ITHit.WebDAV.Server.DavException)"/>.
        /// </param>
        public abstract Task CopyToAsync(
            IItemCollectionAsync destFolder,
            string destName,
            bool deep,
            MultistatusException multistatus);

        /// <summary>
        /// Moves this item to the destination folder under a new name.
        /// </summary>
        /// <param name="destFolder">Destination folder.</param>
        /// <param name="destName">Name of the destination item.</param>
        /// <param name="multistatus">If some items fail to copy but operation in whole shall be continued, add
        /// information about the error into <paramref name="multistatus"/> using 
        /// <see cref="MultistatusException.AddInnerException(string,ITHit.WebDAV.Server.DavException)"/>.
        /// </param>
        public abstract Task MoveToAsync(IItemCollectionAsync destFolder, string destName, MultistatusException multistatus);

        /// <summary>
        /// Deletes this item.
        /// </summary>
        /// <param name="multistatus">If some items fail to delete but operation in whole shall be continued, add
        /// information about the error into <paramref name="multistatus"/> using
        /// <see cref="MultistatusException.AddInnerException(string,ITHit.WebDAV.Server.DavException)"/>.
        /// </param>
        public abstract Task DeleteAsync(MultistatusException multistatus);

        /// <summary>
        /// Retrieves user defined property values.
        /// </summary>
        /// <param name="names">Names of dead properties which values to retrieve.</param>
        /// <param name="allprop">Whether all properties shall be retrieved.</param>
        /// <returns>Property values.</returns>
        public async Task<IEnumerable<PropertyValue>> GetPropertiesAsync(IList<PropertyName> names, bool allprop)
        {
            if (allprop)
            {
                // get all properties
                return await getPropertyValuesAsync();
            }

            //Get requested property values.
            var propertyValues = await getPropertyValuesAsync();
            return propertyValues.Where(property => names.Contains(property.QualifiedName));
        }

        /// <summary>
        /// Retrieves names of all user defined properties.
        /// </summary>
        /// <returns>Property names.</returns>
        public async Task<IEnumerable<PropertyName>> GetPropertyNamesAsync()
        {
            var propertyValues = await getPropertyValuesAsync();
            return propertyValues.Select(p => p.QualifiedName);
        }

        /// <summary>
        /// Saves dead property values to NTFS alternate data stream.
        /// </summary>
        /// <param name="setProps">Properties to be set.</param>
        /// <param name="delProps">Properties to be deleted.</param>
        /// <param name="multistatus">We don't use it here.</param>
        public virtual async Task UpdatePropertiesAsync(
            IList<PropertyValue> setProps,
            IList<PropertyName> delProps,
            MultistatusException multistatus)
        {
            foreach (PropertyValue propToSet in setProps)
            {
                // Microsoft Mini-redirector may update file creation date, modification date and access time passing properties:
                // <Win32CreationTime xmlns="urn:schemas-microsoft-com:">Thu, 28 Mar 2013 20:15:34 GMT</Win32CreationTime>
                // <Win32LastModifiedTime xmlns="urn:schemas-microsoft-com:">Thu, 28 Mar 2013 20:36:24 GMT</Win32LastModifiedTime>
                // <Win32LastAccessTime xmlns="urn:schemas-microsoft-com:">Thu, 28 Mar 2013 20:36:24 GMT</Win32LastAccessTime>
                // In this case update creation and modified date in your storage or do not save this properties at all, otherwise 
                // Windows Explorer will display creation and modification date from this props and it will differ from the values 
                // in the Created and Modified fields in your storage 
                if (propToSet.QualifiedName.Namespace == "urn:schemas-microsoft-com:")
                {
                    if (propToSet.QualifiedName.Name == "Win32CreationTime")
                    {
                        fileSystemInfo.CreationTimeUtc =
                            DateTime.Parse(propToSet.Value, new System.Globalization.CultureInfo("en-US")).ToUniversalTime();
                    }
                    else if (propToSet.QualifiedName.Name == "Win32LastModifiedTime")
                    {
                        fileSystemInfo.LastWriteTimeUtc =
                            DateTime.Parse(propToSet.Value, new System.Globalization.CultureInfo("en-US")).ToUniversalTime();
                    }
                }
                else
                {
                    var propertyValues = await getPropertyValuesAsync();
                    var existingProp = propertyValues.FirstOrDefault(p => p.QualifiedName == propToSet.QualifiedName);

                    if (existingProp != null)
                    {
                        existingProp.Value = propToSet.Value;
                    }
                    else
                    {
                        propertyValues.Add(propToSet);
                    }
                }
            }

            var propertyValuesForRemove = await getPropertyValuesAsync();
            propertyValuesForRemove.RemoveAll(prop => delProps.Contains(prop.QualifiedName));

            await context.FileOperationAsync(
                this,
                () => RewriteStreamAsync("Properties", propertyValues),
                Privilege.Write);

            // You should not update modification date/time here. Mac OS X Finder expects that properties update do not change the file modification date.
        }

        /// <summary>
        /// Returns instance of <see cref="IPrincipalAsync"/> which represents current user.
        /// </summary>
        /// <returns>Current user.</returns>
        /// <remarks>
        /// This method is usually called by the Engine when CalDAV/CardDAV client 
        /// is trying to discover current user URL.
        /// </remarks>
        public async Task<IPrincipalAsync> GetCurrentUserPrincipalAsync()
        {
            // Typically there is no need to load all user properties here, only current 
            // user ID (or name) is required to form the user URL: [DAVLocation]/acl/users/[UserID]
            return AclFactory.GetPrincipalFromSid(context.WindowsIdentity.User.Value, context);
        }

        /// <summary>
        /// Retrieves full local path for this file/folder.
        /// </summary>
        /// <returns>Full local path.</returns>
        internal string GetFullPath()
        {
            return fileSystemInfo.FullName.TrimEnd('\\');
        }

        /// <summary>
        /// Gets NTFS stream with specified name and deserializes object of type <typeparamref name="T"/>
        /// from it.
        /// </summary>
        /// <typeparam name="T">Type of object serialized in stream.</typeparam>
        /// <param name="streamName">Name of NTFS stream stored as NTFS stream in this file/folder.</param>
        /// <param name="logger">Instance of <see cref="logger"/>.</param>
        /// <returns>Deserialized object.</returns>
        internal T GetStreamAndDeserialize<T>(string streamName, ILogger logger) where T : new()
        {
            return context.FileOperation(
                this,
                () => SerializationUtils.GetStreamAndDeserialize<T>(streamName, fileSystemInfo, logger),
                Privilege.Read);
        }

        /// <summary>
        /// Gets NTFS stream with specified name and deserializes object of type <typeparamref name="T"/>
        /// from it.
        /// </summary>
        /// <typeparam name="T">Type of object serialized in stream.</typeparam>
        /// <param name="streamName">Name of NTFS stream stored as NTFS stream in this file/folder.</param>
        /// <param name="logger">Instance of <see cref="logger"/>.</param>
        /// <returns>Deserialized object.</returns>
        internal async Task<T> GetStreamAndDeserializeAsync<T>(string streamName, ILogger logger) where T : new()
        {
            return await context.FileOperationAsync(
                this,
                () => SerializationUtils.GetStreamAndDeserializeAsync<T>(streamName, fileSystemInfo, logger),
                Privilege.Read);
        }

        /// <summary>
        /// Serializes object <paramref name="o"/> and saves it into specified NTFS stream.
        /// </summary>
        /// <typeparam name="T">Type of object to be serialized.</typeparam>
        /// <param name="streamName">NTFS stream name.</param>
        /// <param name="o">Object to be serialized.</param>
        protected abstract Task RewriteStreamAsync<T>(string streamName, T o);

        /// <summary>
        /// Deletes NTFS stream with specified name.
        /// </summary>
        /// <param name="fileName">File name.</param>
        /// <param name="streamName">NTFS stream name.</param>
        protected abstract void DeleteStream(string fileName, string streamName);
   
     
        /// <summary>
        /// Retrieves list of user defined propeties for this item.
        /// </summary>
        /// <returns>List of user defined properties.</returns>
        private async Task<List<PropertyValue>> getPropertyValuesAsync()
        {
            if (propertyValues == null)
            {
                // Deserialize property values from alternate data stream.
                propertyValues = await GetStreamAndDeserializeAsync<List<PropertyValue>>("Properties", context.Logger);
            }

            return propertyValues;
        }
    }
}
