using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;

using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Class1;

namespace WebDAVServer.Class1Listener
{
    /// <summary>
    /// Base class for items like files, folders, versions etc.
    /// </summary>
    public abstract class DavHierarchyItem : IHierarchyItemAsync
    {
        protected DavContext Context { get; private set; }

        public Guid ItemId { get; private set; }

        public string Name { get; private set; }

        public string Path { get; private set; }

        public DateTime Created { get; private set; }

        public DateTime Modified { get; private set; }

        protected Guid ParentId { get; private set; }

        public DavHierarchyItem(
            DavContext context,
            Guid itemId,
            Guid parentId,
            string name,
            string path,
            DateTime created,
            DateTime modified)
        {
            Context = context;
            ItemId = itemId;
            ParentId = parentId;
            Name = name;
            Path = path;
            Created = created;
            Modified = modified;
        }

        public async Task<DavFolder> GetParentAsync()
        {
            string[] parts = Path.TrimEnd('/').Split('/');
            string parentParentPath = "/";
            if (parts.Length > 2)
            {
                parentParentPath = string.Join("/", parts, 0, parts.Length - 2);
            }

            string command =
                @"SELECT 
                     ItemID
                   , ParentItemId
                   , ItemType
                   , Name
                   , Created
                   , Modified
                                      , FileAttributes                  FROM Item
                  WHERE ItemId = @ItemId";

            var davFolders = await Context.ExecuteItemAsync<DavFolder>(
                parentParentPath,
                command,
                "@ItemId", ParentId);
            return davFolders.FirstOrDefault();
        }

        public abstract Task CopyToAsync(
            IItemCollectionAsync destFolder,
            string destName,
            bool deep,
            MultistatusException multistatus);

        public abstract Task MoveToAsync(IItemCollectionAsync destFolder, string destName, MultistatusException multistatus);

        public abstract Task DeleteAsync(MultistatusException multistatus);

        public async Task<IEnumerable<PropertyValue>> GetPropertiesAsync(IList<PropertyName> names, bool allprop)
        {
            IList<PropertyValue> propVals = await Context.ExecutePropertyValueAsync(
                    "SELECT Name, Namespace, PropVal FROM Property WHERE ItemID = @ItemID",
                    "@ItemID", ItemId);

            if (allprop)
            {
                return propVals;
            }
            else
            {
                IList<PropertyValue> requestedPropVals = new List<PropertyValue>();
                foreach (PropertyValue p in propVals)
                {
                    if (names.Contains(p.QualifiedName))
                    {
                        requestedPropVals.Add(p);
                    }
                }
                return requestedPropVals;
            }
        }

        public virtual async Task UpdatePropertiesAsync(
            IList<PropertyValue> setProps,
            IList<PropertyName> delProps,
            MultistatusException multistatus)
        {

            foreach (PropertyValue p in setProps)
            {
                // Microsoft Mini-redirector may update file creation date, modification date and access time passing properties:
                // <Win32CreationTime xmlns="urn:schemas-microsoft-com:">Thu, 28 Mar 2013 20:15:34 GMT</Win32CreationTime>
                // <Win32LastModifiedTime xmlns="urn:schemas-microsoft-com:">Thu, 28 Mar 2013 20:36:24 GMT</Win32LastModifiedTime>
                // <Win32LastAccessTime xmlns="urn:schemas-microsoft-com:">Thu, 28 Mar 2013 20:36:24 GMT</Win32LastAccessTime>
                // In this case update creation and modified date in your storage or do not save this properties at all, otherwise 
                // Windows Explorer will display creation and modification date from this props and it will differ from the values 
                // in the Created and Modified fields in your storage 
                if (p.QualifiedName.Namespace == "urn:schemas-microsoft-com:")
                {
                    if (p.QualifiedName.Name == "Win32CreationTime")
                    {
                        await SetDbFieldAsync("Created", DateTime.Parse(p.Value, new System.Globalization.CultureInfo("en-US")).ToUniversalTime());
                    }
                    else if (p.QualifiedName.Name == "Win32LastModifiedTime")
                    {
                        await SetDbFieldAsync("Modified", DateTime.Parse(p.Value, new System.Globalization.CultureInfo("en-US")).ToUniversalTime());
                    }
                }
                else
                {
                    await SetPropertyAsync(p); // create or update property
                }
            }

            foreach (PropertyName p in delProps)
            {
                await RemovePropertyAsync(p.Name, p.Namespace);
            }

            // You should not update modification date/time here. Mac OS X Finder expects that properties update do not change the file modification date.
        }

        public async Task<IEnumerable<PropertyName>> GetPropertyNamesAsync()
        {
            IList<PropertyName> propNames = new List<PropertyName>();
            foreach (PropertyValue propName in await GetPropertiesAsync(new PropertyName[0], true))
            {
                propNames.Add(propName.QualifiedName);
            }
            return propNames;
        }

        protected async Task SetPropertyAsync(PropertyValue prop)
        {
            string selectCommand =
                @"SELECT Count(*) FROM Property
                  WHERE ItemID = @ItemID AND Name = @Name AND Namespace = @Namespace";

            int count = await Context.ExecuteScalarAsync<int>(
                selectCommand,
                "@ItemID", ItemId,
                "@Name", prop.QualifiedName.Name,
                "@Namespace", prop.QualifiedName.Namespace);

            // insert
            if (count == 0)
            {
                string insertCommand = @"INSERT INTO Property(ItemID, Name, Namespace, PropVal)
                                          VALUES(@ItemID, @Name, @Namespace, @PropVal)";

                await Context.ExecuteNonQueryAsync(
                    insertCommand,
                    "@PropVal", prop.Value,
                    "@ItemID", ItemId,
                    "@Name", prop.QualifiedName.Name,
                    "@Namespace", prop.QualifiedName.Namespace);
            }
            else
            {
                // update
                string command = @"UPDATE Property
                      SET PropVal = @PropVal
                      WHERE ItemID = @ItemID AND Name = @Name AND Namespace = @Namespace";

                await Context.ExecuteNonQueryAsync(
                    command,
                    "@PropVal", prop.Value,
                    "@ItemID", ItemId,
                    "@Name", prop.QualifiedName.Name,
                    "@Namespace", prop.QualifiedName.Namespace);
            }
        }

        protected async Task RemovePropertyAsync(string name, string ns)
        {
            string command = @"DELETE FROM Property
                              WHERE ItemID = @ItemID
                              AND Name = @Name
                              AND Namespace = @Namespace";

            await Context.ExecuteNonQueryAsync(
                command,
                "@ItemID", ItemId,
                "@Name", name,
                "@Namespace", ns);
        }

        internal async Task<DavFolder> CopyThisItemAsync(DavFolder destFolder, DavHierarchyItem destItem, string destName)
        {
            // returns created folder, if any, otherwise null
            DavFolder createdFolder = null;

            Guid destID;
            if (destItem == null)
            {
                // copy item
                string commandText =
                    @"INSERT INTO Item(
                           ItemId
                         , Name
                         , Created
                         , Modified
                         , ParentItemId
                         , ItemType
                         , Content
                         , ContentType
                         , SerialNumber
                         )
                      SELECT
                           @Identity
                         , @Name
                         , GETUTCDATE()
                         , GETUTCDATE()
                         , @Parent
                         , ItemType
                         , Content
                         , ContentType
                         , SerialNumber
                      FROM Item
                      WHERE ItemId = @ItemId";

                destID = Guid.NewGuid();
                await Context.ExecuteNonQueryAsync(
                    commandText,
                    "@Name", destName,
                    "@Parent", destFolder.ItemId,
                    "@ItemId", ItemId,
                    "@Identity", destID);

                await destFolder.UpdateModifiedAsync();

                if (this is IFolderAsync)
                {
                    createdFolder = new DavFolder(
                        Context,
                        destID,
                        destFolder.ItemId,
                        destName,
                        destFolder.Path + EncodeUtil.EncodeUrlPart(destName) + "/",
                        DateTime.UtcNow,
                        DateTime.UtcNow);
                }
            }
            else
            {
                // update existing destination
                destID = destItem.ItemId;

                string commandText = @"UPDATE Item SET
                                       Modified = GETUTCDATE()
                                       , ItemType = src.ItemType
                                       , ContentType = src.ContentType
                                       FROM (SELECT * FROM Item WHERE ItemId=@SrcID) src
                                       WHERE Item.ItemId=@DestID";

                await Context.ExecuteNonQueryAsync(
                    commandText,
                    "@SrcID", ItemId,
                    "@DestID", destID);

                // remove old properties from the destination
                await Context.ExecuteNonQueryAsync(
                    "DELETE FROM Property WHERE ItemID = @ItemID",
                    "@ItemID", destID);
            }

            // copy properties
            string command =
                @"INSERT INTO Property(ItemID, Name, Namespace, PropVal)
                  SELECT @DestID, Name, Namespace, PropVal
                  FROM Property
                  WHERE ItemID = @SrcID";

            await Context.ExecuteNonQueryAsync(
                command,
                "@SrcID", ItemId,
                "@DestID", destID);

            return createdFolder;
        }

        internal async Task MoveThisItemAsync(DavFolder destFolder, string destName, DavFolder parent)
        {
            string command =
                @"UPDATE Item SET
                     Name = @Name,
                     ParentItemId = @Parent
                  WHERE ItemId = @ItemID";

            await Context.ExecuteNonQueryAsync(
                command,
                "@ItemID", ItemId,
                "@Name", destName,
                "@Parent", destFolder.ItemId);

            await parent.UpdateModifiedAsync();
            await destFolder.UpdateModifiedAsync();
        }

        internal async Task DeleteThisItemAsync()
        {
            await DeleteThisItemAsync(await GetParentAsync());
        }

        internal async Task DeleteThisItemAsync(DavFolder parentFolder)
        {
            await Context.ExecuteNonQueryAsync(
                "DELETE FROM Item WHERE ItemId = @ItemID",
                "@ItemID", ItemId);

            if (parentFolder != null)
            {
                await parentFolder.UpdateModifiedAsync();
            }
        }

        internal async Task UpdateModifiedAsync()
        {
            await Context.ExecuteNonQueryAsync(
                "UPDATE Item SET Modified = GETUTCDATE() WHERE ItemId = @ItemId",
                "@ItemId", ItemId);
        }

        protected string CurrentUserName
        {
            get { return Context.User != null ? Context.User.Identity.Name : string.Empty; }
        }        

        protected void SetDbField<T>(string columnName, T value)
        {
            string commandText = string.Format("UPDATE Item SET {0} = @value WHERE ItemId = @ItemId", columnName);
            Context.ExecuteNonQuery(
                commandText,
                "@value", value,
                "@ItemId", ItemId);
        }
        protected async Task SetDbFieldAsync<T>(string columnName, T value)
        {
            string commandText = string.Format("UPDATE Item SET {0} = @value WHERE ItemId = @ItemId", columnName);
            await Context.ExecuteNonQueryAsync(
                commandText,
                "@value", value,
                "@ItemId", ItemId);
        }
    }
}
