using System;
using System.DirectoryServices.AccountManagement;
using System.IO;
using System.Linq;
using System.Net;
using System.Runtime.InteropServices;
using System.Security.Principal;
using System.Web;
using System.Configuration;
using System.Threading.Tasks;

using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Acl;
using ITHit.WebDAV.Server.Quota;
using WebDAVServer.NtfsNtlmAcl.Acl;

namespace WebDAVServer.NtfsNtlmAcl
{
    /// <summary>
    /// Implementation of <see cref="DavContext"/>.
    /// Resolves hierarchy items by paths.
    /// </summary>
    public class DavContext : DavContextBaseAsync, IDisposable
    {
        /// <summary>
        /// Path to folder where files are stored.
        /// </summary>
        private readonly string repositoryPath;

        /// <summary>
        /// Instance of <see cref="ILogger"/> to be used for logging.
        /// </summary>
        private readonly ILogger logger;

        /// <summary>
        /// Disk full windows error code.
        /// </summary>
        private const int ERROR_DISK_FULL = 0x70;

        /// <summary>
        /// Currently logged in identity.
        /// </summary>
        private readonly IIdentity identity;

        /// <summary>
        /// A <see cref="PrincipalContext"/> to be used for windows users operations.
        /// </summary>
        private PrincipalContext principalContext;

        /// <summary>
        /// Initializes a new instance of the DavContext class.
        /// </summary>
        /// <param name="httpContext"><see cref="HttpContext"/> instance.</param>
        public DavContext(HttpContext httpContext) : base(httpContext)
        {
            logger = WebDAVServer.NtfsNtlmAcl.Logger.Instance;
            string configRepositoryPath = (ConfigurationManager.AppSettings["RepositoryPath"] ?? string.Empty).TrimEnd('\\');
            repositoryPath = configRepositoryPath.StartsWith("~") ?
                HttpContext.Current.Server.MapPath(configRepositoryPath) : configRepositoryPath;

            if (!Directory.Exists(repositoryPath))
            {
                WebDAVServer.NtfsNtlmAcl.Logger.Instance.LogError("Repository path specified in Web.config is invalid.", null);
            }

            if (httpContext.User != null)
                identity = httpContext.User.Identity;
        }

        /// <summary>
        /// Gets repository path.
        /// </summary>
        public string RepositoryPath
        {
            get { return repositoryPath; }
        }

        /// <summary>
        /// Gets <see cref="ILogger"/> instance.
        /// </summary>
        public ILogger Logger
        {
            get { return logger; }
        }

        /// <summary>
        /// Gets user name.
        /// </summary>
        /// <remarks>In case of windows authentication returns user name without domain part.</remarks>
        public string UserName
        {
            get
            {
                int i = identity.Name.IndexOf("\\");
                return i > 0 ? identity.Name.Substring(i + 1, identity.Name.Length - i - 1) : identity.Name;
            }
        }

        /// <summary>
        /// Gets currently authenticated user.
        /// </summary>
        public WindowsIdentity WindowsIdentity
        {
            get
            {
                WindowsIdentity windowsIdentity = identity as WindowsIdentity;
                if (windowsIdentity != null && !windowsIdentity.IsAnonymous)
                {
                    return windowsIdentity;
                }

                if (AnonymousUser != null)
                {
                    return AnonymousUser;
                }

                throw new Exception("Anonymous user is not configured.");
            }
        }

        /// <summary>
        /// Gets domain of currently logged in user.
        /// </summary>
        public string Domain
        {
            get
            {
                int i = WindowsIdentity.Name.IndexOf("\\");
                return i > 0 ? WindowsIdentity.Name.Substring(0, i) : Environment.MachineName;
            }
        }

        /// <summary>
        /// Gets or sets user configured as anonymous.
        /// </summary>
        public WindowsIdentity AnonymousUser { get; set; }

        /// <summary>
        /// Retrieves or creates <see cref="PrincipalContext"/> to be used for user related operations.
        /// </summary>
        /// <returns>Instance of <see cref="PrincipalContext"/>.</returns>
        public PrincipalContext GetPrincipalContext()
        {
            if (principalContext == null)
            {
                if (string.IsNullOrEmpty(Domain) ||
                    Environment.MachineName.Equals(Domain, StringComparison.InvariantCultureIgnoreCase))
                {
                    principalContext = new PrincipalContext(ContextType.Machine, Domain);
                }
                else
                {
                   principalContext = new PrincipalContext(ContextType.Domain);
                }
            }

            return principalContext;
        }

        /// <summary>
        /// Performs operation which creates, deletes or modifies user or group.
        /// Performs impersonification and exception handling.
        /// </summary>
        /// <param name="action">Action which performs action with a user or group.</param>
        public void PrincipalOperation(Action action)
        {
            using (impersonate())
            {
                try
                {
                    action();
                }
                catch (PrincipalOperationException ex)
                {
                    Logger.LogError("Principal operation failed", ex);
                    throw new DavException("Principal operation failed", ex);
                }
            }
        }

        /// <summary>
        /// Performs operation which queries, creates, deletes or modifies user or group.
        /// Performs impersonification and exception handling.
        /// </summary>
        /// <param name="func">Function to perform.</param>
        /// <typeparam name="T">Type of operation result.</typeparam>
        /// <returns>The value which <paramref name="func"/> returned.</returns>
        public T PrincipalOperation<T>(Func<T> func)
        {
            using (impersonate())
            {
                try
                {
                    return func();
                }
                catch (PrincipalOperationException ex)
                {
                    Logger.LogError("Principal operation failed", ex);
                    throw new DavException("Principal operation failed", ex);
                }
            }
        }

        /// <summary>
        /// Creates <see cref="IHierarchyItemAsync"/> instance by path.
        /// </summary>
        /// <param name="path">Item relative path including query string.</param>
        /// <returns>Instance of corresponding <see cref="IHierarchyItemAsync"/> or null if item is not found.</returns>
        public override async Task<IHierarchyItemAsync> GetHierarchyItemAsync(string path)
        {
            path = path.Trim(new[] { ' ', '/' });

            //remove query string.
            int ind = path.IndexOf('?');
            if (ind > -1)
            {
                path = path.Remove(ind);
            }

            IHierarchyItemAsync item = null;

            // Return items from [DAVLocation]/acl/ folder and subfolders.
            item = await AclFactory.GetAclItemAsync(this, path);
            if (item != null)
                return item;

            item = DavFolder.GetFolder(this, path);
            if (item != null)
                return item;

            item = DavFile.GetFile(this, path);
            if (item != null)
                return item;

            logger.LogDebug("Could not find item that corresponds to path: " + path);

            return null; // no hierarchy item that corresponds to path parameter was found in the repository
        }

        /// <summary>
        /// Returns the physical file path that corresponds to the specified virtual path on the Web server.
        /// </summary>
        /// <param name="path">Path relative to WebDAV root folder.</param>
        /// <returns>Corresponding path in file system.</returns>
        internal string MapPath(string path)
        {
            //Convert to local file system path by decoding every part, reversing slashes and appending
            //to repository root.
            string[] encodedParts = path.Split(new[] { "/" }, StringSplitOptions.RemoveEmptyEntries);
            string[] decodedParts = encodedParts.Select<string, string>(EncodeUtil.DecodeUrlPart).ToArray();
            string fileSystemPath = Path.Combine(repositoryPath, string.Join("\\", decodedParts));
            return fileSystemPath;
        }

        /// <summary>
        /// Performs file system operation with translating exceptions to those expected by WebDAV engine.
        /// </summary>
        /// <param name="item">Item on which operation is performed.</param>
        /// <param name="action">The action to be performed.</param>
        /// <param name="privilege">Privilege which is needed to perform the operation. If <see cref="UnauthorizedAccessException"/> is thrown
        /// this method will convert it to <see cref="NeedPrivilegesException"/> exception and specify this privilege in it.</param>
        public void FileOperation(IHierarchyItemAsync item, Action action, Privilege privilege)
        {
            try
            {
                using (impersonate())
                {
                    action();
                }
            }
            catch (UnauthorizedAccessException)
            {
                var ex = new NeedPrivilegesException("Not enough privileges");
                ex.AddRequiredPrivilege(item.Path, privilege);
                throw ex;
            }
            catch (IOException ex)
            {
                int hr = Marshal.GetHRForException(ex);
                if (hr == ERROR_DISK_FULL)
                {
                    throw new InsufficientStorageException();
                }

                throw new DavException(ex.Message, DavStatus.CONFLICT);
            }
        }

        /// <summary>
        /// Performs file system operation with translating exceptions to those expected by WebDAV engine.
        /// </summary>
        /// <param name="item">Item on which operation is performed.</param>
        /// <param name="action">The action to be performed.</param>
        /// <param name="privilege">Privilege which is needed to perform the operation. If <see cref="UnauthorizedAccessException"/> is thrown
        /// this method will convert it to <see cref="NeedPrivilegesException"/> exception and specify this privilege in it.</param>
        public async Task FileOperationAsync(IHierarchyItemAsync item, Func<Task> actionAsync, Privilege privilege)
        {
            try
            {
                using (impersonate())
                {
                   await actionAsync();
                }
            }
            catch (UnauthorizedAccessException)
            {
                var ex = new NeedPrivilegesException("Not enough privileges");
                ex.AddRequiredPrivilege(item.Path, privilege);
                throw ex;
            }
            catch (IOException ex)
            {
                int hr = Marshal.GetHRForException(ex);
                if (hr == ERROR_DISK_FULL)
                {
                    throw new InsufficientStorageException();
                }

                throw new DavException(ex.Message, DavStatus.CONFLICT);
            }
        }

        /// <summary>
        /// Performs file system operation with translating exceptions to those expected by WebDAV engine.
        /// </summary>
        /// <param name="item">Item on which operation is performed.</param>
        /// <param name="func">The action to be performed.</param>
        /// <param name="privilege">Privilege which is needed to perform the operation.
        /// If <see cref="UnauthorizedAccessException"/> is thrown  this method will convert it to
        /// <see cref="NeedPrivilegesException"/> exception and specify this privilege in it.</param>
        /// <typeparam name="T">Type of operation's result.</typeparam>
        /// <returns>Result returned by <paramref name="func"/>.</returns>
        public async Task<T> FileOperationAsync<T>(IHierarchyItemAsync item, Func<Task<T>> actionAsync, Privilege privilege)
        {
            try
            {
                using (impersonate())
                {
                    return await actionAsync();
                }
            }
            catch (UnauthorizedAccessException)
            {
                var ex = new NeedPrivilegesException("Not enough privileges");
                ex.AddRequiredPrivilege(item.Path, privilege);
                throw ex;
            }
            catch (IOException ex)
            {
                int hr = Marshal.GetHRForException(ex);
                if (hr == ERROR_DISK_FULL)
                {
                    throw new InsufficientStorageException();
                }

                throw new DavException(ex.Message, DavStatus.CONFLICT);
            }
        }

        /// <summary>
        /// Performs file system operation with translating exceptions to those expected by WebDAV engine, except
        /// <see cref="UnauthorizedAccessException"/> which must be caught and translated manually.
        /// </summary>        
        /// <param name="action">The action to be performed.</param>
        public async Task FileOperationAsync(Func<Task> actionAsync)
        {
            try
            {
                using (impersonate())
                {
                    await actionAsync();
                }
            }
            catch (IOException ex)
            {
                int hr = Marshal.GetHRForException(ex);
                if (hr == ERROR_DISK_FULL)
                {
                    throw new InsufficientStorageException();
                }

                throw new DavException(ex.Message, DavStatus.CONFLICT);
            }
        }

        /// <summary>
        /// Performs file system operation with translating exceptions to those expected by WebDAV engine, except
        /// <see cref="UnauthorizedAccessException"/> which must be caught and translated manually.
        /// </summary>        
        /// <param name="action">The action to be performed.</param>
        public void FileOperation(Action action)
        {
            try
            {
                using (impersonate())
                {
                    action();
                }
            }
            catch (IOException ex)
            {
                int hr = Marshal.GetHRForException(ex);
                if (hr == ERROR_DISK_FULL)
                {
                    throw new InsufficientStorageException();
                }

                throw new DavException(ex.Message, DavStatus.CONFLICT);
            }
        }

        /// <summary>
        /// Performs file system operation with translating exceptions to those expected by WebDAV engine.
        /// </summary>
        /// <param name="item">Item on which operation is performed.</param>
        /// <param name="func">The action to be performed.</param>
        /// <param name="privilege">Privilege which is needed to perform the operation.
        /// If <see cref="UnauthorizedAccessException"/> is thrown  this method will convert it to
        /// <see cref="NeedPrivilegesException"/> exception and specify this privilege in it.</param>
        /// <typeparam name="T">Type of operation's result.</typeparam>
        /// <returns>Result returned by <paramref name="func"/>.</returns>
        public T FileOperation<T>(IHierarchyItemAsync item, Func<T> func, Privilege privilege)
        {
            try
            {
                using (impersonate())
                {
                    return func();
                }
            }
            catch (UnauthorizedAccessException)
            {
                var ex = new NeedPrivilegesException("Not enough privileges");
                ex.AddRequiredPrivilege(item.Path, privilege);
                throw ex;
            }
            catch (IOException ex)
            {
                int hr = Marshal.GetHRForException(ex);
                if (hr == ERROR_DISK_FULL)
                {
                    throw new InsufficientStorageException();
                }

                throw new DavException(ex.Message, DavStatus.CONFLICT);
            }
        }

        /// <summary>
        /// Dispose everything we have.
        /// </summary>
        public void Dispose()
        {
            if (principalContext != null)
            {
                principalContext.Dispose();
            }
        }

        /// <summary>
        /// Impersonates current user.
        /// </summary>
        /// <returns>Impersonation context, which must be disposed to 'unimpersonate'</returns>
        private WindowsImpersonationContext impersonate()
        {
            return LogonUtil.DuplicateToken(WindowsIdentity).Impersonate();
        }
    }
}
