
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading.Tasks;

using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Acl;
using ITHit.WebDAV.Server.Class1;
using WebDAVServer.NtfsNtlmAcl.Acl;
using WebDAVServer.NtfsNtlmAcl.NtfsStreamsEngine;
using ITHit.WebDAV.Server.Search;

namespace WebDAVServer.NtfsNtlmAcl
{
    /// <summary>
    /// Folder in WebDAV repository.
    /// </summary>
    public class DavFolder : DavHierarchyItem, IFolderAsync
    {

        /// <summary>
        /// Corresponding instance of <see cref="DirectoryInfo"/>.
        /// </summary>
        private readonly DirectoryInfo dirInfo;

        /// <summary>
        /// Returns folder that corresponds to path.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/></param>
        /// <param name="path">Encoded path relative to WebDAV root.</param>
        /// <returns>DavFolder instance or null if physical folder not found in file system.</returns>
        public static DavFolder GetFolder(DavContext context, string path)
        {
            string folderPath = context.MapPath(path).TrimEnd('\\');
            DirectoryInfo folder = new DirectoryInfo(folderPath);
            // to block vulnerability when "%20" folder can be injected into path and folder.Exists returns 'true'
            if (!folder.Exists || String.Compare(folder.FullName.TrimEnd('\\'), folderPath, StringComparison.OrdinalIgnoreCase) != 0)
                return null;

            return new DavFolder(folder, context, path);
        }

        /// <summary>
        /// Initializes a new instance of the DavFolder class.
        /// </summary>
        /// <param name="directory">Instance of <see cref="DirectoryInfo"/> class with information about the folder in file system.</param>
        /// <param name="context">Instance of <see cref="DavContext"/>.</param>
        /// <param name="path">Relative to WebDAV root folder path.</param>
        protected DavFolder(DirectoryInfo directory, DavContext context, string path)
            : base(directory, context, path.TrimEnd('/') + "/")
        {
            dirInfo = directory;
        }

        /// <summary>
        /// Retrieves children of this folder.
        /// </summary>
        /// <param name="propNames">List of properties to retrieve with the children. They will be queried by the engine later.</param>
        /// <returns>Children of the folder.</returns>
        public virtual async Task<IEnumerable<IHierarchyItemAsync>> GetChildrenAsync(IList<PropertyName> propNames)
        {
            // Enumerates all child files and folders.
            // You can filter children items in this implementation and 
            // return only items that you want to be visible for this 
            // particular user.

            IList<IHierarchyItemAsync> children = new List<IHierarchyItemAsync>();

            FileSystemInfo[] fileInfos =
                context.FileOperation(
                    this,
                    () => dirInfo.GetFileSystemInfos(),
                    Privilege.Read);

            foreach (FileSystemInfo fileInfo in fileInfos)
            {
                string childPath = Path + EncodeUtil.EncodeUrlPart(fileInfo.Name);
                IHierarchyItemAsync child = await context.GetHierarchyItemAsync(childPath);
                if (child != null)
                {
                    children.Add(child);
                }
            }

            return children;
        }

        /// <summary>
        /// Creates file with name <paramref name="name"/> in this folder.
        /// </summary>
        /// <param name="name">Name of the file.</param>
        /// <returns>The file just created.</returns>
        public async Task<IFileAsync> CreateFileAsync(string name)
        {
            await RequireHasTokenAsync();
            string fileName = System.IO.Path.Combine(fileSystemInfo.FullName, name);
            
            context.FileOperation(
                this,
                () => new FileStream(fileName, FileMode.CreateNew).Dispose(),
                Privilege.Bind);

            return (IFileAsync) await context.GetHierarchyItemAsync(Path + EncodeUtil.EncodeUrlPart(name));
        }

        /// <summary>
        /// Create folder with specified name in the folder.
        /// </summary>
        /// <param name="name">Name of the folder.</param>
        virtual public async Task CreateFolderAsync(string name)
        {
            await RequireHasTokenAsync();
            context.FileOperation(
                this,
                () => dirInfo.CreateSubdirectory(name),
                Privilege.Bind);
        }

        /// <summary>
        /// Delete this item.
        /// </summary>
        /// <param name="multistatus"><see cref="MultistatusException"/> to populate with child files and folders failed to delete.</param>
        public override async Task DeleteAsync(MultistatusException multistatus)
        {
            if (await GetParentAsync() == null)
            {
                throw new DavException("Cannot delete root.", DavStatus.NOT_ALLOWED);
            }
            await RequireHasTokenAsync();
            bool allChildrenDeleted = true;
            foreach (IHierarchyItemAsync child in await GetChildrenAsync(new PropertyName[0]))
            {
                try
                {
                    await child.DeleteAsync(multistatus);
                }
                catch (DavException ex)
                {
                    //continue the operation if a child failed to delete. Tell client about it by adding to multistatus.
                    multistatus.AddInnerException(child.Path, ex);
                    allChildrenDeleted = false;
                }
            }

            if (allChildrenDeleted)
            {
                context.FileOperation(
                    this,
                    () => dirInfo.Delete(),
                    Privilege.Unbind);
            }
        }
        
        /// <summary>
        /// Move this folder to folder <paramref name="destFolder"/>.
        /// </summary>
        /// <param name="destFolder">Destination folder.</param>
        /// <param name="destName">Name for this folder at destination.</param>
        /// <param name="multistatus">Instance of <see cref="MultistatusException"/>
        /// to fill with errors ocurred while moving child items.</param>
        public override async Task MoveToAsync(IItemCollectionAsync destFolder, string destName, MultistatusException multistatus)
        {
            await RequireHasTokenAsync();
            var targetFolder = destFolder as DavFolder;
            if (targetFolder == null)
            {
                throw new DavException("Target folder doesn't exist", DavStatus.CONFLICT);
            }

            if (isRecursive(targetFolder))
            {
                throw new DavException("Cannot move folder to its subtree.", DavStatus.FORBIDDEN);
            }

            string newDirPath = System.IO.Path.Combine(targetFolder.GetFullPath(), destName);
            string targetPath = targetFolder.Path + EncodeUtil.EncodeUrlPart(destName);

            try
            {
                //remove item with the same name at destination if it exists.
                IHierarchyItemAsync item = await context.GetHierarchyItemAsync(targetPath);
                if (item != null)
                    await item.DeleteAsync(multistatus);

                await targetFolder.CreateFolderAsync(destName);
            }
            catch (DavException ex)
            {
                //Continue the operation but report error with destination path to client.
                multistatus.AddInnerException(targetPath, ex);
                return;
            }

            //Move child items
            bool movedAll = true;
            IFolderAsync createdFolder = (IFolderAsync) await context.GetHierarchyItemAsync(targetPath);
            foreach (DavHierarchyItem item in await GetChildrenAsync(new PropertyName[0]))
            {
                try
                {
                    await item.MoveToAsync(createdFolder, item.Name, multistatus);
                }
                catch (DavException ex)
                {
                    //continue the operation but report error with child item to client.
                    multistatus.AddInnerException(item.Path, ex);
                    movedAll = false;
                }
            }

            if (movedAll)
            {
                await DeleteAsync(multistatus);
            }
        }

        /// <summary>
        /// Copies the folder to another folder.
        /// </summary>
        /// <param name="destFolder">Destination parent folder.</param>
        /// <param name="destName">Destination folder name.</param>
        /// <param name="deep">Whether child items shall be copied.</param>
        /// <param name="multistatus">Instance of <see cref="MultistatusException"/>
        /// to fill with errors ocurred while moving child items.</param>
        public override async Task CopyToAsync(
            IItemCollectionAsync destFolder,
            string destName,
            bool deep,
            MultistatusException multistatus)
        {
            var targetFolder = destFolder as DavFolder;
            if (targetFolder == null)
            {
                throw new DavException("Target folder doesn't exist", DavStatus.CONFLICT);
            }

            if (isRecursive(targetFolder))
            {
                throw new DavException("Cannot copy to subfolder", DavStatus.FORBIDDEN);
            }

            string newDirLocalPath = System.IO.Path.Combine(targetFolder.GetFullPath(), destName);
            string targetPath = targetFolder.Path + EncodeUtil.EncodeUrlPart(destName);

            // Create folder at destination.
            try
            {
                if (!Directory.Exists(newDirLocalPath))
                {
                    await targetFolder.CreateFolderAsync(destName);
                }
            }
            catch (DavException ex)
            {
                //continue, but report error to client for the target item.
                multistatus.AddInnerException(targetPath, ex);
            }

            //Copy children
            IFolderAsync createdFolder = (IFolderAsync) await context.GetHierarchyItemAsync(targetPath);
            foreach (DavHierarchyItem item in await GetChildrenAsync(new PropertyName[0]))
            {
                if (!deep && item is DavFolder)
                {
                    continue;
                }

                try
                {
                    await item.CopyToAsync(createdFolder, item.Name, deep, multistatus);
                }
                catch (DavException ex)
                {
                    // If a child item failed to copy we continue but report error to client.
                    multistatus.AddInnerException(item.Path, ex);
                }
            }
        }

        /// <summary>
        /// Find all resources in the subtree which have Owner equal to current user or group current user belongs to. 
        /// </summary>
        /// <param name="matchBy">Identifies whether to search by owner or group.</param>
        /// <param name="props">Properties requested for found items.</param>
        /// <returns>Matching resources.</returns>
        public override async Task<IEnumerable<IAclHierarchyItemAsync>> GetItemsByPropertyAsync(MatchBy matchBy, IList<PropertyName> props)
        {
            //We support only searching by owner(searching by Group is not supported).
            IList<IAclHierarchyItemAsync> items = new List<IAclHierarchyItemAsync>();
            if (matchBy != MatchBy.Owner)
            {
                return items;
            }

            var currentUser = context.WindowsIdentity;

            foreach (var child in await GetChildrenAsync(props))
            {
                if (!(child is DavHierarchyItem))
                {
                    continue;
                }

                //return this item if it has matching owner.
                DavHierarchyItem hItem = child as DavHierarchyItem;
                var o = (PrincipalBase)await hItem.GetOwnerAsync();
                if (o.Principal.Sid == currentUser.User)
                {
                    items.Add((DavHierarchyItem)child);
                }
                else
                if ((await o.GetGroupMembersAsync()).Any(m => ((PrincipalBase)m).Sid == currentUser.User))
                {
                    items.Add((DavHierarchyItem)child);
                }
                else
                //look for matching child items.
                if (child is DavFolder)
                {
                    DavFolder childDavFolder = child as DavFolder;
                    foreach (var c in await childDavFolder.GetItemsByPropertyAsync(matchBy, props))
                    {
                        items.Add(c); 
                    }
                }
            }
            return items;
        }

        /// <summary>
        /// Serializes object <paramref name="o"/> and saves it into specified NTFS stream.
        /// </summary>
        /// <typeparam name="T">Type of object to be serialized.</typeparam>
        /// <param name="streamName">NTFS stream name.</param>
        /// <param name="o">Object to be serialized.</param>
        protected override async Task RewriteStreamAsync<T>(string streamName, T o)
        {
            await context.FileOperationAsync(
                 this,
                 () => SerializationUtils.RewriteStreamAsync(
                     streamName,
                     fileSystemInfo, o != null ? SerializationUtils.Serialize(o) : null),
                 Privilege.Write);
        }

        /// <summary>
        /// Deletes NTFS stream with specified name.
        /// </summary>
        /// <param name="fileName">File name.</param>
        /// <param name="streamName">NTFS stream name.</param>
        protected override void DeleteStream(string fileName, string streamName)
        {
            new AlternateDataStreamInfo(fileName, streamName).Delete();
        }

        /// <summary>
        /// Determines whether <paramref name="destFolder"/> is inside this folder.
        /// </summary>
        /// <param name="destFolder">Folder to check.</param>
        /// <returns>Returns <c>true</c> if <paramref name="destFolder"/> is inside thid folder.</returns>
        private bool isRecursive(DavFolder destFolder)
        {
            return destFolder.Path.StartsWith(Path);
        }
    }
}
