//!<#@ template language="C#" #>
//!<#@ samplevars Processor="DirectiveProcessor" #>
//!<# if (!Params.CalDav)  return string.Empty; #>
//!<# if (Params.Storage != StorageType.Db)  return string.Empty; #>
using System;
using System.Linq;
using System.IO;
using System.Threading.Tasks;
using System.Collections.Generic;
using System.Data;
using System.Data.SqlClient;
using System.Diagnostics;
//!<# if (Params.VisualStudioVersionNumber < 14) { #>
using System.Linq.Expressions;
using System.Reflection;
//!<# } #>


using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.CalDav;

using ITHit.Collab;
using ITHit.Collab.Calendar;


namespace ITHit.WebDav.Server.SampleModel.__CalDavSql__.CalDav
{
    /// <summary>
    /// Represents calendar file. Every clendar file stores an event or to-do that consists of one or 
    /// more event / to-do components and time zones descriptions.
    /// 
    /// Instances of this class correspond to the following path: [DAVLocation]/calendars/[CalendarFolderID]/[UID].ics
    /// </summary>
    /// <example>
    /// 
    /// +--- Calendar file [UID1].ics ---------+
    /// |                                      |
    /// |  +-- Time zone component ---------+  | -- Time zones are are not stored in DB, they are generated automatically 
    /// |  | TZID: Zone X                   |  |    during serialization, based on TZIDs found in event or to-do.
    /// |  | ...                            |  |
    /// |  +--------------------------------+  |
    /// |                                      |
    /// |  +-- Time zone component ---------+  |
    /// |  | TZID: Zone Y                   |  | -- Time zone IDs could be either IANA (Olson) IDs or System (Windows) IDs.
    /// |  | ...                            |  |
    /// |  +--------------------------------+  |
    /// |  ...                                 |
    /// |                                      |
    /// |                                      |
    /// |  +-- Event component -------------+  | -- Event / do-do components are stored in [EventComponent] table.
    /// |  | UID: [UID1]                    |  |    
    /// |  | RRULE: FREQ=DAILY              |  | 
    /// |  | SUMMARY: Event A               |  |
    /// |  | ...                            |  |
    /// |  +--------------------------------+  |
    /// |                                      |
    /// |  +-- Event component -------------+  | -- In case of recurring events/to-dos there could be more than one component
    /// |  | UID: [UID1]                    |  |    per file. All event/to-do components within a single calendar file share
    /// |  | RECURRENCE-ID: 20151028        |  |    the same UID but have different RECURRENCE-IDs. 
    /// |  | SUMMARY: Instance 5 of Event A |  |    
    /// |  | ...                            |  |    iOS / OS X UIDs are case sensitive (uppercase GUIDs).
    /// |  +--------------------------------+  |    Bynari WebDAV Collaborator for MS Outlook UIDs are over 100 chars in length.
    /// |  ...                                 |
    /// |                                      |
    /// |                                      |
    /// +--------------------------------------+
    /// 
    /// 
    /// 
    ///    +-- Event component -------------+
    ///    |                                |
    ///    | UID: [UID1]                    | 
    ///    | SUMMARY: Event A               |
    ///    | START: 20151016T080000         |
    ///    | RRULE: FREQ=DAILY              |
    ///    | ...                            |
    ///    |                                |
    ///    | EXDATE: 20151018T080000        | -- Recurrence exception dates are stored in [RecurrenceException] table.
    ///    | EXDATE: 20151020T080000        |
    ///    | ...                            |
    ///    |                                |
    ///    | ATTENDEE: mail1@server.com     | -- Attendees are stored in [Attendee] table.
    ///    | ATTENDEE: mail2@srvr.com       |
    ///    | ...                            |
    ///    |                                |
    ///    | ATTACH: /9j/4VGuf+Sw...        | -- Attachments are stored in [Attachment] table.
    ///    | ATTACH: https://serv/file.docx |
    ///    | ...                            |
    ///    |                                |
    ///    |  +-- Alarm Component -------+  | -- Alarms are stored in [Alarm] table.
    ///    |  | ACTION: DISPLAY          |  |
    ///    |  | ...                      |  |
    ///    |  +--------------------------+  |
    ///    |                                |
    ///    |  +-- Alarm Component -------+  |
    ///    |  | ACTION: EMAIL            |  |
    ///    |  | ...                      |  |
    ///    |  +--------------------------+  |
    ///    |  ...                           |
    ///    |                                |
    ///    +--------------------------------+
    /// 
    /// </example>
    public class CalendarFile : DavHierarchyItem, ICalendarFileAsync
    {
        /// <summary>
        /// Calendar file extension.
        /// </summary>
        public static string Extension = ".ics";

        /// <summary>
        /// Loads calendar files contained in a calendar folder by calendar folder ID.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/> class.</param>
        /// <param name="calendarFolderId">Calendar for which events or to-dos should be loaded.</param>
        /// <param name="propsToLoad">Specifies which properties should be loaded.</param>
        /// <returns>List of <see cref="ICalendarFileAsync"/> items.</returns>
        public static async Task<IEnumerable<ICalendarFileAsync>> LoadByCalendarFolderIdAsync(DavContext context, Guid calendarFolderId, PropsToLoad propsToLoad)
        {
            // propsToLoad == PropsToLoad.Minimum -> Typical GetChildren call by iOS, Android, eM Client, etc CalDAV clients
            // [Summary] is typically not required in GetChildren call, 
            // they are extracted for demo purposes only, to be displayed in Ajax File Browser.

            // propsToLoad == PropsToLoad.All -> Bynari call, it requires all props in GetChildren call.

            if (propsToLoad != PropsToLoad.Minimum)
                throw new NotImplementedException("LoadByCalendarFolderIdAsync is implemented only with PropsToLoad.Minimum.");

            string sql = @"SELECT  *        FROM [CalendarFile]   WHERE [CalendarFolderId] = @CalendarFolderId
                          ;SELECT [Summary] FROM [EventComponent] WHERE [UID] IN 
                                   (SELECT [UID] FROM [CalendarFile] WHERE [CalendarFolderId] = @CalendarFolderId)";

            sql = string.Format(sql, GetScPropsToLoad(propsToLoad));
            

            return await LoadAsync(context, sql, "@CalendarFolderId", calendarFolderId);
        }

        /// <summary>
        /// Loads calendar files by list of UIDs.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/> class.</param>
        /// <param name="uids">File UIDs to load.</param>
        /// <param name="propsToLoad">Specifies which properties should be loaded.</param>
        /// <returns>List of <see cref="ICalendarFileAsync"/> items.</returns>
        public static async Task<IEnumerable<ICalendarFileAsync>> LoadByUidsAsync(DavContext context, IEnumerable<string> uids, PropsToLoad propsToLoad)
        {
            // Get IN clause part with list of file UIDs for SELECT.
            string selectIn = string.Join(", ", uids.Select(a => string.Format("'{0}'", a)).ToArray());

            string sql = @"SELECT * FROM [CalendarFile] WHERE [UID] IN ({1})";

            if(propsToLoad==PropsToLoad.All)
            {
                // Here we do not select attachments content because it could be very large,
                // we only set [ContentExists] flag marking that it should be loaded during IContent.ReadAsync call.
                sql += @";SELECT * FROM [EventComponent]      WHERE [UID] IN ({1})
                         ;SELECT * FROM [RecurrenceException] WHERE [UID] IN ({1})
                         ;SELECT * FROM [Alarm]               WHERE [UID] IN ({1})
                         ;SELECT * FROM [Attendee]            WHERE [UID] IN ({1})
                         ;SELECT [EventComponentId], [UID], [MediaType], [ExternalUrl], 
                             (CASE WHEN [Content] IS NULL THEN 0 ELSE 1 END) AS [ContentExists] 
                                   FROM [Attachment]          WHERE [UID] IN ({1})";
            }

            sql = string.Format(sql, GetScPropsToLoad(propsToLoad), selectIn);

            return await LoadAsync(context, sql);
        }

        /// <summary>
        /// Loads calendar files by SQL.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/> class.</param>
        /// <param name="sql">SQL that queries [CalendarFile], [EventComponent], etc tables.</param>
        /// <param name="prms">List of SQL parameters.</param>
        /// <returns>List of <see cref="ICalendarFileAsync"/> items.</returns>
        private static async Task<IEnumerable<ICalendarFileAsync>> LoadAsync(DavContext context, string sql, params object[] prms)
        {
            IList<ICalendarFileAsync> items = new List<ICalendarFileAsync>();

            Stopwatch stopWatch = Stopwatch.StartNew();

            using (SqlDataReader reader = await context.ExecuteReaderAsync(sql, prms))
            {
                DataTable calendarFiles = new DataTable();
                calendarFiles.Load(reader);

                DataTable eventComponents = new DataTable();
                if (!reader.IsClosed)
                    eventComponents.Load(reader);

                DataTable recurrenceExceptions = new DataTable();
                if (!reader.IsClosed)
                    recurrenceExceptions.Load(reader);

                DataTable alarms = new DataTable();
                if (!reader.IsClosed)
                    alarms.Load(reader);

                DataTable attendees = new DataTable();
                if (!reader.IsClosed)
                    attendees.Load(reader);

                DataTable attachments = new DataTable();
                if (!reader.IsClosed)
                    attachments.Load(reader);


                stopWatch.Stop();
                context.Engine.Logger.LogDebug(string.Format("SQL took: {0}ms", stopWatch.ElapsedMilliseconds));


                foreach (DataRow rowCalendarFile in calendarFiles.Rows)
                {
                    DataRow[] rowsEventComponents      = new DataRow[0];
                    DataRow[] rowsRecurrenceExceptions = new DataRow[0];
                    DataRow[] rowsAlarms               = new DataRow[0];
                    DataRow[] rowsAttendees            = new DataRow[0];
                    DataRow[] rowsAttachments          = new DataRow[0];

                    string uid = rowCalendarFile.Field<string>("UID");

                    if (eventComponents.Columns["UID"] != null)
                    {
                        string filter = string.Format("UID = '{0}'", uid);

                        rowsEventComponents      = eventComponents.Select(filter);
                        rowsRecurrenceExceptions = recurrenceExceptions.Select(filter);
                        rowsAlarms               = alarms.Select(filter);
                        rowsAttendees            = attendees.Select(filter);
                        rowsAttachments          = attachments.Select(filter);
                    }
                    
                    items.Add(new CalendarFile(context, uid, rowCalendarFile, rowsEventComponents, rowsRecurrenceExceptions, rowsAlarms, rowsAttendees, rowsAttachments));
                }
            }

            return items;
        }

        private static string GetScPropsToLoad(PropsToLoad propsToLoad)
        {
            switch (propsToLoad)
            {
                case PropsToLoad.None:
                    return "[UID]";
                case PropsToLoad.Minimum:
                    // [Summary] is typically not required in GetChildren call, 
                    // they are extracted for demo purposes only, to be displayed in Ajax File Browser as a file display name.
                    return "[Summary]";
                case PropsToLoad.All:
                    return "*";
            }
            throw new Exception("Should never come here.");
        }

        /// <summary>
        /// Creates new calendar file. The actual new [CalendarFile], [EventComponent], etc. records are inserted into the database during <see cref="WriteAsync"/> method call.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/> class.</param> 
        /// <param name="calendarFolderId">Calendar folder ID to which this calendar file will belong to.</param>
        /// <returns>Instance of <see cref="CalendarFile"/>.</returns>
        public static CalendarFile CreateCalendarFile(DavContext context, Guid calendarFolderId)
        {
            CalendarFile calendarFile = new CalendarFile(context, null, null, null, null, null, null, null);
            calendarFile.calendarFolderId = calendarFolderId;
            return calendarFile;
        }

        /// <summary>
        /// This file UID.
        /// </summary>
        private readonly string uid = null;

        /// <summary>
        /// Contains data from [CalendarFile] table.
        /// </summary>
        private readonly DataRow rowCalendarFile = null;

        /// <summary>
        /// Contains event or to-do components from [EventComponent] table.
        /// </summary>
        private readonly DataRow[] rowsEventComponents = null;

        /// <summary>
        /// Contains recurrence days exceptions for this event or to-do from from [RecurrenceException] table.
        /// </summary>
        private readonly DataRow[] rowsRecurrenceExceptions = null;

        /// <summary>
        /// Contains alarms for this event or to-do from from [Alarm] table.
        /// </summary>
        private readonly DataRow[] rowsAlarms = null;

        /// <summary>
        /// Contains attendees for this event or to-do from from [Attendee] table.
        /// </summary>
        private readonly DataRow[] rowsAttendees = null;

        /// <summary>
        /// Contains attachments for this event or to-do from from [Attachment] table. 
        /// The [Attachment].[Content] field is never populated in this property because it could be very large. 
        /// Instead the [Attachment].[Content] is read in ReadAsync implementation to reduce memory consumprion.
        /// </summary>
        private readonly DataRow[] rowsAttachments = null;

        /// <summary>
        /// Indicates if this is a newly cleated event/to-do.
        /// </summary>
        private bool isNew
        {
            get { return calendarFolderId != Guid.Empty; }
        }

        /// <summary>
        /// Calendar ID in which the new event or to-do will be created.
        /// </summary>
        private Guid calendarFolderId = Guid.Empty;

        /// <summary>
        /// Gets display name of the event or to-do. Used for demo purposes only, to be displayed in Ajax File Browser.
        /// </summary>
        /// <remarks>CalDAV clients typically never request this property.</remarks>
        public override string Name
        {
            get
            {
                // Show all components summaries contained in this file.
                return string.Join(", ", rowsEventComponents.Select(x => string.Format("[{0}]", x.Field<string>("Summary"))).ToArray());
            }
        }

        /// <summary>
        /// Gets item path.
        /// </summary>
        /// <remarks>[DAVLocation]/calendars/[CalendarFolderID]/[UID].ics</remarks>
        public override string Path
        {
            get
            {
                Guid calendarFolderId = rowCalendarFile.Field<Guid>("CalendarFolderId");
                string uid              = rowCalendarFile.Field<string>("UID");
                return string.Format("{0}{1}/{2}{3}", CalendarsRootFolder.CalendarsRootFolderPath, calendarFolderId, uid, Extension);
            }
        }

        /// <summary>
        /// Gets eTag. Used for synchronization with client application. ETag must change every time the event/to-do is updated.
        /// </summary>
        public string Etag
        {
            get
            {
                byte[] bETag = rowCalendarFile.Field<byte[]>("ETag");
                return BitConverter.ToUInt64(bETag.Reverse().ToArray(), 0).ToString(); // convert timestamp value to number
            }
        }

        /// <summary>
        /// Gets item creation date. Must be in UTC.
        /// </summary>
        public override DateTime Created
        {
            get { return rowCalendarFile.Field<DateTime>("CreatedUtc"); }
        }

        /// <summary>
        /// Gets item modification date. Must be in UTC.
        /// </summary>
        public override DateTime Modified
        {
            get { return rowCalendarFile.Field<DateTime>("ModifiedUtc"); }
        }

        /// <summary>
        /// File content length. Typicaly never requested by CalDAV clients.
        /// </summary>
        /// <remarks>
        /// If -1 is returned the chunked response will be generated if possible. The getcontentlength property will not be generated.
        /// </remarks>
        public long ContentLength
        {
            get { return -1; }
        }

        /// <summary>
        /// File Mime-type/Content-Type.
        /// </summary>
        public string ContentType
        {
	        get { return "text/calendar"; }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CalendarFile"/> class from database source. Used when listing folder content and in multi-get.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/></param>
        /// <param name="uid">Calendar file UID.</param>
        /// <param name="rowCalendarFile">Calendar file info from [CalendarFile] table.</param>
        /// <param name="rowsEventComponents">List of event components for this event or to-do data from [EventComponent] table.</param>
        /// <param name="rowsRecurrenceExceptions">List of recurrence days exceptions for this event or to-do from from [RecurrenceExceptions] table.</param>
        /// <param name="rowsAlarms">List of alarms for this event or to-do from from [Alarm] table.</param>
        /// <param name="rowsAttendees">List of attendees for this event or to-do from from [Attendee] table.</param>
        /// <param name="rowsAttachments">
        /// List of attachments for this event or to-do from from [Attachment] table. 
        /// The [Attachment].[Content] field shoud be never populated in this property - it could be very large. 
        /// Instead the [Attachment].[Content] is read in ReadAsync implementation to reduce memory consumprion.
        /// </param>
        private CalendarFile(DavContext context, string uid, DataRow rowCalendarFile, DataRow[] rowsEventComponents, DataRow[] rowsRecurrenceExceptions, DataRow[] rowsAlarms, DataRow[] rowsAttendees, DataRow[] rowsAttachments)
            : base(context)
        {
            this.uid                      = uid;
            this.rowCalendarFile          = rowCalendarFile;
            this.rowsEventComponents      = rowsEventComponents;
            this.rowsRecurrenceExceptions = rowsRecurrenceExceptions;
            this.rowsAlarms               = rowsAlarms;
            this.rowsAttendees            = rowsAttendees;
            this.rowsAttachments          = rowsAttachments;
        }

        //$<ICalendarFileAsync.WriteAsync
        /// <summary>
        /// Called when event or to-do is being saved to back-end storage.
        /// </summary>
        /// <param name="stream">Stream containing VCALENDAR, typically with a single VEVENT ot VTODO component.</param>
        /// <param name="contentType">Content type.</param>
        /// <param name="startIndex">Starting byte in target file
        /// for which data comes in <paramref name="content"/> stream.</param>
        /// <param name="totalFileSize">Size of file as it will be after all parts are uploaded. -1 if unknown (in case of chunked upload).</param>
        /// <returns>Whether the whole stream has been written.</returns>
        public async Task<bool> WriteAsync(Stream stream, string contentType, long startIndex, long totalFileSize)
        {
            //!<# if (Params.Hosting == HostingEnvironment.IIS) { #>
            //Set timeout to maximum value to be able to upload iCalendar files with large file attachments.
            System.Web.HttpContext.Current.Server.ScriptTimeout = int.MaxValue;

            //!<# } #>
            string iCalendar;
            using (StreamReader reader = new StreamReader(stream))
            {
                iCalendar = reader.ReadToEnd();
            }

            Context.Engine.Logger.LogDebug(iCalendar);

            // Typically the stream contains a single iCalendar that contains one or more event or to-do components.
            IEnumerable<IComponent> calendars = new vFormatter().Deserialize(iCalendar);
            ICalendar2 calendar = calendars.First() as ICalendar2;

            IEnumerable<IEventBase> components = calendar.Events.Cast<IEventBase>();
            if (!components.Any())
            {
                components = calendar.ToDos.Cast<IEventBase>();
            }

            if (components == null)
                throw new DavException("Event or to-do was expected in the input stream, no events or to-dos were found.", DavStatus.UNSUPPORTED_MEDIA_TYPE);

            // All components inside calendar file has the same UID which is equal to file name.
            string uid = components.First().Uid.Text;

            // Save data to [CalendarFile] table.
            await WriteCalendarFileAsync(Context, uid, calendarFolderId, isNew);

            foreach (IEventBase component in components)
            {
                Guid eventComponentId = Guid.NewGuid();

                // Save data to [EventComponent] table.
                await WriteEventComponentAsync(Context, component, eventComponentId, uid);

                // Save recurrence days exceptions for recurring events and to-dos. 
                await WriteRecurrenceExceptionsAsync(Context, component.ExceptionDateTimes, eventComponentId, uid);

                // Save alarms.
                await WriteAlarmsAsync(Context, component.Alarms, eventComponentId, uid);

                // Save attengees.
                await WriteAttendeesAsync(Context, component.Attendees, eventComponentId, uid);

                // Save attachments.
                await WriteAttachmentsAsync(Context, component.Attachments, eventComponentId, uid);
            }

            // Notify attendees that event is created or modified.
            calendar.Method = calendar.CreateMethodProp(MethodType.Request);
            await iMipEventSchedulingTransport.NotifyAttendeesAsync(Context, calendar);

            return true;
        }

        /// <summary>
        /// Saves data to [CalendarFile] table.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/> class.</param>
        /// <param name="uid">File UID to be updated or created.</param>
        /// <param name="calendarFolderId">Calendar folder that contains this file.</param>
        /// <param name="isNew">Flag indicating if this is a new file or file should be updated.</param>
        /// <remarks>
        /// This function deletes records in [EventComponent], [RecurrenceException], [Alarms],
        /// [Attendees] and [Attachments] tables if the event or to-do should be updated.
        /// </remarks>
        private static async Task WriteCalendarFileAsync(DavContext context, string uid, Guid calendarFolderId, bool isNew)
        {
            string sql;
            if (isNew)
            {
                sql =
                    @"INSERT INTO [CalendarFile] (
                          [UID]
                        , [CalendarFolderId]
                    ) VALUES (
                          @UID
                        , @CalendarFolderId
                    )";
            }
            else
            {
                // We can only update record in [CalendarFile] table.
                // There is no way to update [EventComponent], [RecurrenceException], [Alarms], [Attendees] and 
                // [Attachments] for existing event, we must delete all records for this UID and recreate.

                // [ModifiedUtc] field update triggers [ETag] field update which is used for synchronyzation.
                sql =
                    @"UPDATE [CalendarFile] SET 
                        [ModifiedUtc]       = @ModifiedUtc
                    WHERE 
                        [UID] = @UID
                    
                    ; DELETE FROM [EventComponent] WHERE [UID] = @UID;
                    ; DELETE FROM [RecurrenceException] WHERE [UID] = @UID;
                    ; DELETE FROM [Alarm]               WHERE [UID] = @UID;
                    ; DELETE FROM [Attendee]            WHERE [UID] = @UID;
                    ; DELETE FROM [Attachment]          WHERE [UID] = @UID";
            }

            await context.ExecuteNonQueryAsync(sql,
                  "@UID", uid
                , "@CalendarFolderId", calendarFolderId
                , "@ModifiedUtc", DateTime.UtcNow
                );
        }

        /// <summary>
        /// Saves data to [EventComponent] table.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/> class.</param>
        /// <param name="sc">Event or to-do component to read data from.</param>
        /// <param name="eventComponentId">New event component ID.</param>
        /// <param name="uid">File UID.</param>
        private static async Task WriteEventComponentAsync(DavContext context, IEventBase sc, Guid eventComponentId, string uid)
        {
            string sql =
                    @"INSERT INTO [EventComponent] (
                          [EventComponentId]
                        , [UID]
                        , [ComponentType]
                        , [DateTimeStampUtc]
                        , [CreatedUtc]
                        , [LastModifiedUtc]
                        , [Summary]
                        , [Description]
                        , [OrganizerEmail]
                        , [OrganizerCommonName]
                        , [Start]
                        , [StartTimeZoneId]
                        , [End]
                        , [EndTimeZoneId]
                        , [Duration]
                        , [AllDay]
                        , [Class]
                        , [Location]
                        , [Priority]
                        , [Sequence]
                        , [Status]
                        , [Categories]
                        , [RecurFrequency]
                        , [RecurInterval]
                        , [RecurUntil]
                        , [RecurCount]
                        , [RecurWeekStart]
                        , [RecurByDay]
                        , [RecurByMonthDay]
                        , [RecurByMonth]
                        , [RecurBySetPos]
                        , [RecurrenceIdDate]
                        , [RecurrenceIdTimeZoneId]
                        , [RecurrenceIdThisAndFuture]
                        , [EventTransparency]
                        , [ToDoCompletedUtc]
                        , [ToDoPercentComplete]
                    ) VALUES (
                          @EventComponentId
                        , @UID
                        , @ComponentType
                        , @DateTimeStampUtc 
                        , @CreatedUtc
                        , @LastModifiedUtc
                        , @Summary
                        , @Description
                        , @OrganizerEmail
                        , @OrganizerCommonName
                        , @Start, @StartTimeZoneId
                        , @End, @EndTimeZoneId
                        , @Duration
                        , @AllDay
                        , @Class
                        , @Location
                        , @Priority
                        , @Sequence
                        , @Status
                        , @Categories
                        , @RecurFrequency
                        , @RecurInterval
                        , @RecurUntil
                        , @RecurCount
                        , @RecurWeekStart
                        , @RecurByDay
                        , @RecurByMonthDay
                        , @RecurByMonth
                        , @RecurBySetPos
                        , @RecurrenceIdDate
                        , @RecurrenceIdTimeZoneId
                        , @RecurrenceIdThisAndFuture
                        , @EventTransparency
                        , @ToDoCompletedUtc
                        , @ToDoPercentComplete
                    )";


            bool isEvent = sc is IEvent;

            // Get END in case of event or DUE in case of to-do component. 
            ICalDate endProp = isEvent ? (sc as IEvent).End : (sc as IToDo).Due;
            
            await context.ExecuteNonQueryAsync(sql,
                  "@EventComponentId"       , eventComponentId
                , "@UID"                    , uid                                                                   // UID value
                , "@ComponentType"          , isEvent
                , "@DateTimeStampUtc"       , sc.DateTimeStampUtc?.Value?.DateVal                                   // DTSTAMP value
                , "@CreatedUtc"             , sc.CreatedUtc?.Value?.DateVal                                         // CREATED value
                , "@LastModifiedUtc"        , sc.LastModifiedUtc?.Value?.DateVal                                    // LAST-MODIFIED value
                , "@Summary"                , sc.Summary?.Text                                                      // SUMMARY value
                , "@Description"            , sc.Description?.Text                                                  // DESCRIPTION value
                , "@OrganizerEmail"         , sc.Organizer?.Uri?.Replace("mailto:", "")                             // ORGANIZER value
                , "@OrganizerCommonName"    , sc.Organizer?.CommonName                                              // ORGANIZER CN param
                , "@Start"                  , sc.Start?.Value?.DateVal                                              // DTSTART value
                , "@StartTimeZoneId"        , sc.Start?.Value?.DateVal.Kind == DateTimeKind.Utc ? TimeZoneInfo.Utc.Id : sc.Start?.TimeZoneId  // DTSTART TZID param
                , "@End"                    , endProp?.Value?.DateVal                                               // DTEND or DUE value
                , "@EndTimeZoneId"          , endProp?.Value?.DateVal.Kind == DateTimeKind.Utc ? TimeZoneInfo.Utc.Id : endProp?.TimeZoneId    // DTEND or DUE TZID param
                , "@Duration"               , sc.Duration?.Value?.Ticks                                             // DURATION value
                , "@AllDay"                 , !sc.Start?.Value?.Components.HasFlag(DateComponents.Time) // Check if start contains the time part to determine if this is a all-day event/to-do.
                , "@Class"                  , sc.Class?.Value.Name                                                  // CLASS value
                , "@Location"               , sc.Location?.Text                                                     // LOCATION value
                , "@Priority"               , sc.Priority?.Value                                                    // PRIORITY value
                , "@Sequence"               , sc.Sequence?.Value                                                    // SEQUENCE value
                , "@Status"                 , sc.Status?.Value.Name                                                 // STATUS value
                , "@Categories"             , ListToString<string>(sc.Categories.Select(x => ListToString<string>(x.Categories, ",")), ";") // CATEGORIES value
                , "@RecurFrequency"         , sc.RecurrenceRule?.Frequency?.ToString()                              // RRULE FREQ value part
                , "@RecurInterval"          , (int?)sc.RecurrenceRule?.Interval                                     // RRULE INTERVAL value part
                , "@RecurUntil"             , sc.RecurrenceRule?.Until?.DateVal                                     // RRULE UNTIL value part
                , "@RecurCount"             , (int?)sc.RecurrenceRule?.Count                                        // RRULE COUNT value part
                , "@RecurWeekStart"         , sc.RecurrenceRule?.WeekStart?.ToString()                              // RRULE WKST value part
                , "@RecurByDay"             , ListToString<DayRule>(sc.RecurrenceRule?.ByDay)                       // RRULE BYDAY value part
                , "@RecurByMonthDay"        , ListToString<short>(sc.RecurrenceRule?.ByMonthDay)                    // RRULE BYMONTHDAY value part
                , "@RecurByMonth"           , ListToString<ushort>(sc.RecurrenceRule?.ByMonth)                      // RRULE BYMONTH value part
                , "@RecurBySetPos"          , ListToString<short>(sc.RecurrenceRule?.BySetPos)                      // RRULE BYSETPOS value part
                , "@RecurrenceIdDate"       , sc.RecurrenceId?.Value.DateVal                                        // RECURRENCE-ID value
                , "@RecurrenceIdTimeZoneId" , sc.RecurrenceId?.TimeZoneId                                           // RECURRENCE-ID TZID param
                , "@RecurrenceIdThisAndFuture", sc.RecurrenceId?.IsThisAndFuture                                    // RECURRENCE-ID RANGE param
                , "@EventTransparency"      , (sc as IEvent)?.Transparency?.IsTransparent                           // VEVENT TRANSP value
                , "@ToDoCompletedUtc"       , (sc as IToDo)?.CompletedUtc?.Value?.DateVal                           // VTODO COMPLETED value
                , "@ToDoPercentComplete"    , (sc as IToDo)?.PercentComplete?.Value                                 // VTODO PERCENT-COMPLETE value
                );
        }

        /// <summary>
        /// Converts <see cref="IEnumerable{T}"/> to string. Returns null if the list is empty.
        /// </summary>
        /// <returns>String that contains elements separated by ','.</returns>
        private static string ListToString<T>(IEnumerable<T> arr, string separator = ",")
        {
            if ((arr == null) || !arr.Any())
                return null;
            //!<# if (Params.TargetFramework == "3.5") { #>
            return string.Join(separator, arr.Select(obj => obj.ToString()).ToArray());
            //!<# } else { #>
            return string.Join<T>(separator, arr);
            //!<# } #>
        }

        /// <summary>
        /// Saves data to [RecurrenceException] table.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/> class.</param>
        /// <param name="recurrenceExceptions">Event or to-do recurrence exceptions dates to be saved.</param>
        /// <param name="eventComponentId">Event component to associate these recurrence exceptions dates with.</param>
        /// <param name="uid">File UID.</param>
        private static async Task WriteRecurrenceExceptionsAsync(DavContext context, IPropertyList<ICalDateList> recurrenceExceptions, Guid eventComponentId, string uid)
        {
            // Typically CalDAV clients pass a single date value per EXDATE property.

            string sql =
                @"INSERT INTO [RecurrenceException] (
                      [EventComponentId]
                    , [UID]
                    , [ExceptionDate]
                    , [TimeZoneId]
                    , [AllDay]
                ) VALUES {0}";

            List<string> valuesSql = new List<string>();
            List<object> parameters = new List<object>(new object[] {
                  "@EventComponentId", eventComponentId
                , "@UID", uid
            });

            int i = 0;
            foreach (ICalDateList dateListProp in recurrenceExceptions)
            {
                foreach (Date date in dateListProp.Dates)
                {
                    if (date == null)
                        continue; // failed fo parse date

                    valuesSql.Add(string.Format(@"(
                      @EventComponentId
                    , @UID
                    , @ExceptionDate{0}
                    , @TimeZoneId{0}
                    , @AllDay{0}
                    )", i));

                    parameters.AddRange(new object[] {
                  //  "@EventComponentId"
                  //  "@UID" added for performance optimization purposes
                      "@ExceptionDate" +i, date.DateVal                                                                         // EXDATE value
                    , "@TimeZoneId"    +i, date.DateVal.Kind == DateTimeKind.Utc ? TimeZoneInfo.Utc.Id : dateListProp.TimeZoneId// EXDATE TZID param
                    , "@AllDay"        +i, !date.Components.HasFlag(DateComponents.Time)                                        // EXDATE DATE or DATE-TIME
                    });
                }
                i++;
            }

            if (i > 0)
            {
                await context.ExecuteNonQueryAsync(string.Format(sql, string.Join(", ", valuesSql.ToArray())), parameters.ToArray());
            }
        }

        /// <summary>
        /// Saves data to [Alarm] table.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/> class.</param>
        /// <param name="alarms">List of alarms to be saved.</param>
        /// <param name="eventComponentId">Event component to associate these alarms with.</param>
        /// <param name="uid">File UID.</param>
        private static async Task WriteAlarmsAsync(DavContext context, IComponentList<IAlarm> alarms, Guid eventComponentId, string uid)
        {
            string sql =
                @"INSERT INTO [Alarm] (
                      [EventComponentId]
                    , [UID]
                    , [Action]
                    , [TriggerAbsoluteDateTimeUtc]
                    , [TriggerRelativeOffset]
                    , [TriggerRelatedStart]
                    , [Summary]
                    , [Description]
                    , [Duration]
                    , [Repeat]
                ) VALUES {0}";

            List<string> valuesSql = new List<string>();
            List<object> parameters = new List<object>(new object[] {
                  "@EventComponentId", eventComponentId
                , "@UID", uid
            });

            int i = 0;
            foreach (IAlarm alarm in alarms)
            {
                valuesSql.Add(string.Format(@"(
                      @EventComponentId
                    , @UID
                    , @Action{0}
                    , @TriggerAbsoluteDateTimeUtc{0}
                    , @TriggerRelativeOffset{0}
                    , @TriggerRelatedStart{0}
                    , @Summary{0}
                    , @Description{0}
                    , @Duration{0}
                    , @Repeat{0}
                    )", i));

                parameters.AddRange(new object[] {
                  //  "@EventComponentId"
                  //  "@UID" added for performance optimization purposes
                      "@Action"                     +i, alarm.Action.Action.Name                                                        // Alarm ACTION property
                    , "@TriggerAbsoluteDateTimeUtc" +i, alarm.Trigger?.AbsoluteDateTimeUtc                                              // Alarm TRIGGER property
                    , "@TriggerRelativeOffset"      +i, alarm.Trigger?.RelativeOffset?.Ticks                                            // Alarm TRIGGER property
                    , "@TriggerRelatedStart"        +i, alarm.Trigger==null ? (bool?)null : alarm.Trigger.Related == RelatedType.Start  // Alarm trigger RELATED param
                    , "@Summary"                    +i, alarm.Summary?.Text                                                             // Alarm SUMMARY property
                    , "@Description"                +i, alarm.Description?.Text                                                         // Alarm DESCRIPTION property
                    , "@Duration"                   +i, alarm.Duration?.Value?.Ticks                                                    // Alarm DURATION property
                    , "@Repeat"                     +i, alarm.Repeat?.Value                                                             // Alarm REPEAT property
                });
                i++;
            }

            if (i > 0)
            {
                await context.ExecuteNonQueryAsync(string.Format(sql, string.Join(", ", valuesSql.ToArray())), parameters.ToArray());
            }
        }

        /// <summary>
        /// Saves data to [Attendee] table.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/> class.</param>
        /// <param name="attendees">List of attendees to be saved.</param>
        /// <param name="eventComponentId">Event component to associate these attendees with.</param>
        /// <param name="uid">File UID.</param>
        private static async Task WriteAttendeesAsync(DavContext context, IPropertyList<IAttendee> attendees, Guid eventComponentId, string uid)
        {
            string sql =
                @"INSERT INTO [Attendee] (
                      [EventComponentId]
                    , [UID]
                    , [Email]
                    , [CommonName]
                    , [DirectoryEntryRef]
                    , [Language]
                    , [UserType]
                    , [SentBy]
                    , [DelegatedFrom]
                    , [DelegatedTo]
                    , [Rsvp]
                    , [ParticipationRole]
                    , [ParticipationStatus]
                ) VALUES {0}";

            List<string> valuesSql = new List<string>();
            List<object> parameters = new List<object>(new object[] {
                  "@EventComponentId", eventComponentId
                , "@UID", uid
            });

            int i = 0;
            foreach (IAttendee attendee in attendees)
            {
                valuesSql.Add(string.Format(@"(
                      @EventComponentId
                    , @UID
                    , @Email{0}
                    , @CommonName{0}
                    , @DirectoryEntryRef{0}
                    , @Language{0}
                    , @UserType{0}
                    , @SentBy{0}
                    , @DelegatedFrom{0}
                    , @DelegatedTo{0}
                    , @Rsvp{0}
                    , @ParticipationRole{0}
                    , @ParticipationStatus{0}
                )", i));

                parameters.AddRange(new object[] {
                  //  "@EventComponentId"
                  //  "@UID" added for performance optimization purposes
                      "@Email"              +i, attendee.Uri?.Replace("mailto:", "")    // Attendee value
                    , "@CommonName"         +i, attendee.CommonName                     // Attendee CN parameter
                    , "@DirectoryEntryRef"  +i, attendee.Dir                            // Attendee DIR parameter
                    , "@Language"           +i, attendee.Language                       // Attendee LANGUAGE parameter
                    , "@UserType"           +i, attendee.UserType?.Name                 // Attendee CUTYPE parameter
                    , "@SentBy"             +i, attendee.SentBy                         // Attendee SENT-BY parameter
                    , "@DelegatedFrom"      +i, attendee.DelegatedFrom.FirstOrDefault() // Attendee DELEGATED-FROM parameter, here we assume only 1 delegator for the sake of simplicity
                    , "@DelegatedTo"        +i, attendee.DelegatedTo.FirstOrDefault()   // Attendee DELEGATED-TO parameter, here we assume only 1 delegatee for the sake of simplicity
                    , "@Rsvp"               +i, attendee.Rsvp == RsvpType.True          // Attendee RSVP parameter
                    , "@ParticipationRole"  +i, attendee.ParticipationRole?.Name        // Attendee ROLE parameter
                    , "@ParticipationStatus"+i, attendee.ParticipationStatus?.Name      // Attendee PARTSTAT parameter
                });
                i++;
            }

            if (i > 0)
            {
                await context.ExecuteNonQueryAsync(string.Format(sql, string.Join(", ", valuesSql.ToArray())), parameters.ToArray());
            }
        }

        /// <summary>
        /// Saves data to [Attachment] table.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/> class.</param>
        /// <param name="attachments">List of attachments to be saved.</param>
        /// <param name="eventComponentId">Event component to associate these attachments with.</param>
        /// <param name="uid">File UID.</param>
        private static async Task WriteAttachmentsAsync(DavContext context, IPropertyList<IMedia> attachments, Guid eventComponentId, string uid)
        {
            // It is recommended to keep attchment size below 256Kb. In case files over 1Mb should 
            // be stored, use SQL FILESTREAM, FileTable or store content in file system.

            string sql =
                @"INSERT INTO [Attachment] (
                      [EventComponentId]
                    , [UID]
                    , [MediaType]
                    , [ExternalUrl]
                    , [Content]
                ) VALUES (
                      @EventComponentId
                    , @UID
                    , @MediaType
                    , @ExternalUrl
                    , @Content
                )";


            foreach (IMedia attachment in attachments)
            {
                // To insert NULL to VARBINARY column SqlParameter must be passed with Size=-1 and Value=DBNull.Value.
                SqlParameter contentPram = new SqlParameter("@Content", SqlDbType.VarBinary, -1);
                contentPram.Value = DBNull.Value;

                if (!attachment.IsExternal)
                {
                    byte[] content = Convert.FromBase64String(attachment.Base64Data);
                    contentPram.Size = content.Length;
                    contentPram.Value = content;
                }

                await context.ExecuteNonQueryAsync(sql,
                      "@EventComponentId", eventComponentId
                    , "@UID"             , uid
                    , "@MediaType"       , attachment.MediaType
                    , "@ExternalUrl"     , attachment.IsExternal ? attachment.Uri : null
                    , contentPram
                    );
            }
        }
        //$>

        /// <summary>
        /// Called when client application deletes this file.
        /// </summary>
        /// <param name="multistatus">Error description if case delate failed. Ignored by most clients.</param>
        public override async Task DeleteAsync(MultistatusException multistatus)
        {
            // Notify attendees that event is canceled if deletion is successful.
            ICalendar2 cal = await GetCalendarAsync();

            await Context.ExecuteNonQueryAsync(
                "DELETE FROM [CalendarFile] WHERE UID=@UID", "@UID", uid);

            cal.Method = cal.CreateMethodProp(MethodType.Cancel);
            await iMipEventSchedulingTransport.NotifyAttendeesAsync(Context, cal);

        }

        //$<ICalendarFileAsync.ReadAsync
        /// <summary>
        /// Called when the event or to-do must be read from back-end storage.
        /// </summary>
        /// <param name="output">Stream to write event or to-do content.</param>
        /// <param name="startIndex">Index to start reading data from back-end storage. Used for segmented reads, not used by CalDAV clients.</param>
        /// <param name="count">Number of bytes to read. Used for segmented reads, not used by CalDAV clients.</param>
        /// <returns></returns>
        public async Task ReadAsync(Stream output, long startIndex, long count)
        {
            ICalendar2 cal = await GetCalendarAsync();
            new vFormatter().Serialize(output, cal);
        }
        
        /// <summary>
        /// Creates calendar based on loaded data rows. Loads attachments from database if required.
        /// </summary>
        /// <returns>Object that implements <see cref="ICalendar2"/>.</returns>
        private async Task<ICalendar2> GetCalendarAsync()
        {
            ICalendar2 cal = CalendarFactory.CreateCalendar2();
            cal.ProductId = cal.CreateTextProp("-//IT Hit//Collab Lib//EN");

            // Recurrent event or to-do can contain more than one VEVENT/VTODO component in one file.
            foreach (DataRow rowEventComponent in rowsEventComponents)
            {

                // add either event or to-do to the calendar
                bool isEvent = rowEventComponent.Field<bool>("ComponentType");
                IEventBase sc;
                if (isEvent)
                {
                    sc = cal.Events.CreateComponent();
                    cal.Events.Add(sc as IEvent);
                }
                else
                {
                    sc = cal.ToDos.CreateComponent();
                    cal.ToDos.Add(sc as IToDo);
                }

                // Read component properties from previously loaded [EventComponent] rows.
                ReadEventComponent(sc, rowEventComponent, cal);


                Guid eventComponentId = rowEventComponent.Field<Guid>("EventComponentId");

                // Get [RecurrenceException] rows that belong to this event component only and read recurrence exceptions dates.
                IEnumerable<DataRow> rowsThisScRecurrenceExceptions = rowsRecurrenceExceptions.Where(x => x.Field<Guid>("EventComponentId") == eventComponentId);
                ReadRecurrenceExceptions(sc.ExceptionDateTimes, rowsThisScRecurrenceExceptions, cal);

                // Get [Alarm] rows that belong to this event component only and read alarms.
                IEnumerable<DataRow> rowsThisScAlarms = rowsAlarms.Where(x => x.Field<Guid>("EventComponentId") == eventComponentId);
                ReadAlarms(sc.Alarms, rowsThisScAlarms, cal);

                // Get [Attendee] rows that belong to this event component only and read attendees.
                IEnumerable<DataRow> rowsThisScAttendees = rowsAttendees.Where(x => x.Field<Guid>("EventComponentId") == eventComponentId);
                ReadAttendees(sc.Attendees, rowsThisScAttendees, cal);

                // Get [Attachment] rows that belong to this event component only.
                // Read attachments, load [Attachment].[Content] if required.
                IEnumerable<DataRow> rowsThisScAttachments = rowsAttachments.Where(x => x.Field<Guid>("EventComponentId") == eventComponentId);
                await ReadAttachmentsAsync(Context, sc.Attachments, rowsThisScAttachments, cal);
            }

            // Generate VTIMEZONE components based on TZID parameters.
            cal.AutoGenerateTimeZones = true;

            return cal;
        }

        /// <summary>
        /// Reads data from [EventComponent] row.
        /// </summary>
        /// <param name="sc">Event or to-do that will be populated from row paramater.</param>
        /// <param name="row">Data from [EventComponent] table to populate sc.</param>
        /// <param name="cal">Calendar object.</param>
        private static void ReadEventComponent(IEventBase sc, DataRow row, ICalendar2 cal)
        {
            bool isAllDay = row.Field<bool?>("AllDay") == true;

            sc.Uid              = cal.CreateTextProp(row.Field<string>("UID"));                                     // UID property, iOS / OS X UID is case sensitive, Bynari WebDAV Collaborator UID is over 100 chars long
            sc.DateTimeStampUtc = cal.CreateDateProp(row.Field<DateTime?>("DateTimeStampUtc"), DateTimeKind.Utc);   // DTSTAMP property
            sc.CreatedUtc       = cal.CreateDateProp(row.Field<DateTime?>("CreatedUtc"), DateTimeKind.Utc);         // CREATED property
            sc.LastModifiedUtc  = cal.CreateDateProp(row.Field<DateTime?>("LastModifiedUtc"), DateTimeKind.Utc);    // LAST-MODIFIED property
            sc.Summary          = cal.CreateCalTextProp(row.Field<string>("Summary"));                              // SUMMARY property
            sc.Description      = cal.CreateCalTextProp(row.Field<string>("Description"));                          // DESCRIPTION property
            sc.Start            = cal.CreateCalDateProp(row.Field<DateTime?>("Start"), row.Field<string>("StartTimeZoneId"), isAllDay);                 // DTSTART property
            sc.Duration         = cal.CreateDurationProp(row.Field<long?>("Duration"));                             // DURATION property
            sc.Class            = cal.CreateClassProp(row.Field<string>("Class"));                                  // CLASS property
            sc.Location         = cal.CreateCalTextProp(row.Field<string>("Location"));                             // LOCATION property
            sc.Priority         = cal.CreateIntegerProp(row.Field<byte?>("Priority"));                              // PRIORITY property
            sc.Sequence         = cal.CreateIntegerProp(row.Field<int?>("Sequence"));                               // SEQ property
            sc.Status           = cal.CreateStatusProp(row.Field<string>("Status"));                                // STATUS property
            sc.Organizer        = cal.CreateCalAddressProp(EmailToUri(row.Field<string>("OrganizerEmail")), row.Field<string>("OrganizerCommonName"));  // ORGANIZER property

            // RECURRENCE-ID property
            sc.RecurrenceId     = cal.CreateRecurrenceIdProp(
                row.Field<DateTime?>("RecurrenceIdDate")                                                            // RECURRENCE-ID value
                , row.Field<string>("RecurrenceIdTimeZoneId")                                                       // RECURRENCE-ID TZID param
                , isAllDay                                                                                          // RECURRENCE-ID DATE or DATE-TIME
                , row.Field<bool?>("RecurrenceIdThisAndFuture")==true);                                             // RECURRENCE-ID RANGE param

            // CATEGORIES property list
            string categories = row.Field<string>("Categories") as string;
            if (!string.IsNullOrEmpty(categories))
            {
                string[] strCatProp = categories.Split(new[] { ';' }, StringSplitOptions.RemoveEmptyEntries);
                foreach (string categoryList in strCatProp)
                {
                    ICategories catProp = sc.Categories.CreateProperty();
                    catProp.Categories = categoryList.Split(new[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
                    sc.Categories.Add(catProp);
                }
            }

            // RRULE property
            string recurFrequency = row.Field<string>("RecurFrequency");
            if (!string.IsNullOrEmpty(recurFrequency))
            {
                sc.RecurrenceRule = cal.CreateProperty<IRecurrenceRule>();

                sc.RecurrenceRule.Frequency = ExtendibleEnum.FromString<FrequencyType>(recurFrequency);             // FREQ rule part
                sc.RecurrenceRule.Interval  = (uint?)row.Field<int?>("RecurInterval");                              // INTERVAL rule part
                sc.RecurrenceRule.Count     = (uint?)row.Field<int?>("RecurCount");                                 // COUNT rule part

                // WKST rule part
                string weekStart = row.Field<string>("RecurWeekStart");
                if (!string.IsNullOrEmpty(weekStart))
                {
                    sc.RecurrenceRule.WeekStart = (DayOfWeek)Enum.Parse(typeof(DayOfWeek), weekStart);
                }

                // UNTIL rule part
                DateTime? until = row.Field<DateTime?>("RecurUntil");
                if (until != null)
                {
                    // UNTIL must be in UTC if DTSTART contains time zone or DTSTART is UTC.
                    // UNTIL must be 'floating' if DTSTART is 'floating'.
                    // UNTIL must be 'all day' if the DTSTART is 'all day'.
                    // https://tools.ietf.org/html/rfc5545#section-3.3.10
                    sc.RecurrenceRule.Until = new Date(
                        DateTime.SpecifyKind(until.Value,
                        sc.Start.Value.DateVal.Kind != DateTimeKind.Local /*floating*/ ? DateTimeKind.Utc : DateTimeKind.Local),
                        sc.Start.Value.Components);
                }

                // BYDAY rule part
                string byDay = row.Field<string>("RecurByDay");
                if (!string.IsNullOrEmpty(byDay))
                {
                    sc.RecurrenceRule.ByDay = byDay.Split(',').Select(x => DayRule.Parse(x)).ToArray();
                }

                // BYMONTHDAY rule part
                string byMonthDay = row.Field<string>("RecurByMonthDay");
                if (!string.IsNullOrEmpty(byMonthDay))
                {
                    sc.RecurrenceRule.ByMonthDay = byMonthDay.Split(',').Select(x => short.Parse(x)).ToArray();
                }

                // BYMONTH rule part
                string byMonth = row.Field<string>("RecurByMonth");
                if (!string.IsNullOrEmpty(byMonth))
                {
                    sc.RecurrenceRule.ByMonth = byMonth.Split(',').Select(x => ushort.Parse(x)).ToArray();
                }

                // BYSETPOS  rule part
                string bySetPos = row.Field<string>("RecurBySetPos");
                if (!string.IsNullOrEmpty(bySetPos))
                {
                    sc.RecurrenceRule.BySetPos = bySetPos.Split(',').Select(x => short.Parse(x)).ToArray();
                }      
            }

            if (sc is IEvent)
            {
                // Properties specific for events only
                IEvent vEvent = sc as IEvent;
                vEvent.End          = cal.CreateCalDateProp(row.Field<DateTime?>("End"), row.Field<string>("EndTimeZoneId"), isAllDay); // DTEND property
                vEvent.Transparency = cal.CreateTransparencyProp(row.Field<bool?>("EventTransparency"));                                // TRANSP property
            }
            else
            {
                // Properties specific for to-dos only
                IToDo vToDo = sc as IToDo;
                vToDo.Due           = cal.CreateCalDateProp(row.Field<DateTime?>("End"), row.Field<string>("EndTimeZoneId"), isAllDay); // DUE property
                vToDo.CompletedUtc  = cal.CreateDateProp(row.Field<DateTime?>("ToDoCompletedUtc"), DateTimeKind.Utc);                   // COMPLETED property
                vToDo.PercentComplete = cal.CreateIntegerProp(row.Field<byte?>("ToDoPercentComplete"));                                 // PERCENT-COMPLETE
            }
        }

        /// <summary>
        /// Reads data from [RecurrenceException] rows.
        /// </summary>
        /// <param name="recurrenceExceptions">Empty recurrence exceptions dates list that will be populated with data from rowsRecurrenceExceptions parameter.</param>
        /// <param name="rowsRecurrenceExceptions">Data from [RecurrenceException] table to populate recurrenceExceptions parameter.</param>
        private static void ReadRecurrenceExceptions(IPropertyList<ICalDateList> recurrenceExceptions, IEnumerable<DataRow> rowsRecurrenceExceptions, ICalendar2 cal)
        {
            foreach (DataRow rowRecurrenceException in rowsRecurrenceExceptions)
            {
                // EXDATE property
                ICalDateList exdate = cal.CreateCalDateListProp(
                    new DateTime[] { rowRecurrenceException.Field<DateTime>("ExceptionDate") }                          // EXDATE value
                    , rowRecurrenceException.Field<string>("TimeZoneId")                                                // EXDATE TZID param
                    , rowRecurrenceException.Field<bool?>("AllDay") == true                                             // EXDATE DATE or DATE-TIME
                    );
                recurrenceExceptions.Add(exdate);
            }
        }

        /// <summary>
        /// Reads data from [Alarm] rows.
        /// </summary>
        /// <param name="alarms">Empty alarms list that will be populated with data from rowsAlarms parameter.</param>
        /// <param name="rowsAlarms">Data from [Alarm] table to populate alarms parameter.</param>
        /// <param name="cal">Calendar object.</param>
        private static void ReadAlarms(IComponentList<IAlarm> alarms, IEnumerable<DataRow> rowsAlarms, ICalendar2 cal)
        {
            foreach (DataRow rowAlarm in rowsAlarms)
            {
                IAlarm alarm = alarms.CreateComponent();                                            // VALARM component

                alarm.Action        = cal.CreateActionProp(rowAlarm.Field<string>("Action"));       // Alarm ACTION property
                alarm.Summary       = cal.CreateCalTextProp(rowAlarm.Field<string>("Summary"));     // Alarm SUMMARY property
                alarm.Description   = cal.CreateCalTextProp(rowAlarm.Field<string>("Description")); // Alarm DESCRIPTION property
                alarm.Duration      = cal.CreateDurationProp(rowAlarm.Field<long?>("Duration"));    // Alarm DURATION property
                alarm.Repeat        = cal.CreateIntegerProp(rowAlarm.Field<int?>("Repeat"));        // Alarm REPEAT property

                // Alarm TRIGGER property
                alarm.Trigger = cal.CreateProperty<ITrigger>();

                DateTime? absolute = rowAlarm.Field<DateTime?>("TriggerAbsoluteDateTimeUtc");
                if (absolute != null)
                {
                    alarm.Trigger.AbsoluteDateTimeUtc = DateTime.SpecifyKind(absolute.Value, DateTimeKind.Utc);
                }

                long? offset = rowAlarm.Field<long?>("TriggerRelativeOffset");
                if (offset != null)
                {
                    alarm.Trigger.RelativeOffset = new TimeSpan(offset.Value);
                }

                // Alarm trigger RELATED param
                bool? related = rowAlarm.Field<bool?>("TriggerRelatedStart");
                if (related != null)
                {
                    alarm.Trigger.Related = related.Value ? RelatedType.Start : RelatedType.End;
                }

                alarms.Add(alarm);
            }
        }

        /// <summary>
        /// Reads data from [Attendee] rows.
        /// </summary>
        /// <param name="attendees">Empty attendees list that will be populated with data from rowsAttendees parameter.</param>
        /// <param name="rowsAttendees">Data from [Attendee] table to populate attendees parameter.</param>
        /// <param name="cal">Calendar object.</param>
        private static void ReadAttendees(IPropertyList<IAttendee> attendees, IEnumerable<DataRow> rowsAttendees, ICalendar2 cal)
        {
            foreach (DataRow rowAttendee in rowsAttendees)
            {                
                IAttendee attendee = attendees.CreateProperty();                                                // ATTENDEE property

                attendee.Uri            = EmailToUri(rowAttendee.Field<string>("Email"));                       // Attendee value
                attendee.CommonName     = rowAttendee.Field<string>("CommonName");                              // Attendee CN parameter
                attendee.Dir            = rowAttendee.Field<string>("DirectoryEntryRef");                       // Attendee DIR parameter
                attendee.Language       = rowAttendee.Field<string>("Language");                                // Attendee LANGUAGE parameter
                attendee.UserType       = StringToEnum<CalendarUserType>(rowAttendee.Field<string>("UserType"));// Attendee CUTYPE parameter
                attendee.SentBy         = EmailToUri(rowAttendee.Field<string>("SentBy"));                      // Attendee SENT-BY parameter
                attendee.DelegatedFrom  = new[] { EmailToUri(rowAttendee.Field<string>("DelegatedFrom")) };     // Attendee DELEGATED-FROM parameter, here we assume only 1 delegator for the sake of simplicity
                attendee.DelegatedTo    = new[] { EmailToUri(rowAttendee.Field<string>("DelegatedTo")) };       // Attendee DELEGATED-TO parameter, here we assume only 1 delegatee for the sake of simplicity

                // Attendee RSVP parameter
                bool? rsvp = rowAttendee.Field<bool?>("Rsvp");
                if (rsvp != null)
                {
                    attendee.Rsvp       = rsvp.Value ? RsvpType.True : RsvpType.False;
                }

                attendee.ParticipationRole   = StringToEnum<ParticipationRoleType>(rowAttendee.Field<string>("ParticipationRole"));     // Attendee ROLE parameter
                attendee.ParticipationStatus = StringToEnum<ParticipationStatusType>(rowAttendee.Field<string>("ParticipationStatus")); // Attendee PARTSTAT parameter

                attendees.Add(attendee);
            }
        }

        /// <summary>
        /// Reads data from [Attachment] rows. Loads [Attachment].[Content] if required.
        /// </summary>
        /// <param name="attachments">Empty attachments list that will be populated with data from rowsAttachments parameter.</param>
        /// <param name="rowsAttachments">Data from [Attachment] table to populate attachments parameter.</param>
        /// <param name="cal">Calendar object.</param>
        private static async Task ReadAttachmentsAsync(DavContext context, IPropertyList<IMedia> attachments, IEnumerable<DataRow> rowsAttachments, ICalendar2 cal)
        {
            // Find if any attachments content should be read from datatbase.
            bool loadContent = rowsAttachments.Any(x => (x.Field<int>("ContentExists") == 1));

            if (loadContent)
            {
                // Reading attachments content from database.
                //!<# if (Params.Hosting == HostingEnvironment.IIS) { #>

                // Set timeout to maximum value to be able to download iCalendar files with large file attachments.
                System.Web.HttpContext.Current.Server.ScriptTimeout = int.MaxValue;
                //!<# } #>
                
                Guid eventComponentId = rowsAttachments.First().Field<Guid>("EventComponentId");
                string sql = "SELECT [MediaType], [ExternalUrl], [Content] FROM [Attachment] WHERE [EventComponentId]=@EventComponentId";

                using (SqlDataReader reader = await context.ExecuteReaderAsync(CommandBehavior.SequentialAccess, sql, "@EventComponentId", eventComponentId))
                {
                    while(await reader.ReadAsync())
                    {
                        IMedia attachment = attachments.CreateProperty();                   // ATTACH property

                        // Attachment FMTTYPE parameter
                        int ordMediaType = reader.GetOrdinal("MediaType");
                        if (!await reader.IsDBNullAsync(ordMediaType))
                        {
                            //!<# if (!Params.TargetFramework45OrLater) { #>
                            attachment.MediaType = reader.GetString(ordMediaType);
                            //!<# } else { #>
                            attachment.MediaType = await reader.GetFieldValueAsync<string>(ordMediaType);
                            //!<# } #>
                        }

                        // Attachment value as URL
                        int ordExternalUrl = reader.GetOrdinal("ExternalUrl");
                        if (!await reader.IsDBNullAsync(ordExternalUrl))
                        {
                            //!<# if (!Params.TargetFramework45OrLater) { #>
                            attachment.Uri = reader.GetString(ordExternalUrl);
                            //!<# } else { #>
                            attachment.Uri = await reader.GetFieldValueAsync<string>(ordExternalUrl);
                            //!<# } #>
                        }
                                     
                        // Attachment value as inline content
                        int ordContent = reader.GetOrdinal("Content");
                        if (!await reader.IsDBNullAsync(ordContent))
                        {
                            //!<# if (!Params.TargetFramework45OrLater) { #>
                            attachment.Base64Data = Convert.ToBase64String((byte[])reader[ordContent]);
                            //!<# } else { #>
                            using (Stream stream = reader.GetStream(ordContent))
                            {
                                using (MemoryStream memory = new MemoryStream())
                                {
                                    await stream.CopyToAsync(memory);
                                    attachment.Base64Data = Convert.ToBase64String(memory.ToArray());
                                }
                            }
                            //!<# } #>
                        }

                        attachments.Add(attachment);
                    }
                }
            }
            else
            {
                // Attachments contain only URLs to external files.
                foreach (DataRow rowAttachment in rowsAttachments)
                {
                    IMedia attachment = attachments.CreateProperty();                   // ATTACH property

                    attachment.MediaType = rowAttachment.Field<string>("MediaType");    // Attachment FMTTYPE parameter
                    attachment.Uri       = rowAttachment.Field<string>("ExternalUrl");  // Attachment value
                    
                    attachments.Add(attachment);
                }
            }
        }

        /// <summary>
        /// Adds "mailto:" schema to e-mail address if "@" is found. If null is passed returns null.
        /// </summary>
        /// <param name="email">E-mail.</param>
        /// <returns>E-mail string with "mailto:" schema.</returns>
        private static string EmailToUri(string email)
        {
            if (email == null)
                return null;

            if (email.IndexOf('@') > 0)
                return string.Format("mailto:{0}", email);

            return email;
        }

        /// <summary>
        /// Converts string to <see cref="ExtendibleEnum"/> of spcified type. Returns <b>null</b> if <b>null</b> is passed. 
        /// If no matching string value is found the <see cref="ExtendibleEnum.Name"/> is set to passed parameter <b>value</b> and <see cref="ExtendibleEnum.Number"/> is set to -1.
        /// </summary>
        /// <typeparam name="T">Type to convert to.</typeparam>
        /// <param name="value">String to convert from.</param>
        /// <returns><see cref="ExtendibleEnum"/> of type <b>T</b> or <b>null</b> if <b>null</b> is passed as a parameter.</returns>
        private static T StringToEnum<T>(string value) where T : ExtendibleEnum, new()
        {
            if (value == null)
                return null;

            T res;
            if (!ExtendibleEnum.TryFromString<T>(value, out res))
            {
                // If no matching value is found create new ExtendibleEnum or type T 
                // with specified string value and default numeric value (-1).
                res = new T();
                res.Name = value;
            }

            return res;
        }
        //$>
        //!<# if (Params.VisualStudioVersionNumber < 14) { #>

        /// <summary>
        /// Analog of Null-conditional operator (?.) for VS 2013 and earlier. Returns null if any property or method 
        /// in expression returns null. Otherwise evaluates expression and returns the result.
        /// </summary>
        /// <remarks>
        /// In case your code is generated by the CalDAV/CardDAV wizard in VS 2015 the null-conditional operator is 
        /// used instead of this function.
        /// </remarks>
        /// <typeparam name="T">Class or struct.</typeparam>
        /// <param name="expression">Linq expression. For example 'Prop1.Prop2.Method1().Method2()'. Any property or method in this expression can return null.</param>
        /// <returns><see cref="object"/> of type <b>T</b> or null.</returns>
        public static object NCO<T>(Expression<Func<T>> expression)
        {
            object membObj = null;
            MethodCallExpression methodCallExpression = null;
            MemberExpression membExpress = null;
            List<MemberExpression> memberExpressions = new List<MemberExpression>();
            if (expression.Body is MethodCallExpression)
            {
                methodCallExpression = expression.Body as MethodCallExpression;
                membExpress = methodCallExpression.Object as MemberExpression;
            }
            else
                membExpress = expression.Body as MemberExpression;

            while (membExpress != null)
            {
                memberExpressions.Add(membExpress);
                membExpress = membExpress.Expression as MemberExpression;
            }

            if (memberExpressions.Count != 0)
            {
                memberExpressions.Reverse();
                MemberExpression lastMembExpr = memberExpressions.First();

                if (Expression.Lambda(lastMembExpr.Expression).Compile().DynamicInvoke() != null)
                    membObj = Expression.Lambda(lastMembExpr).Compile().DynamicInvoke();

                foreach (var membExpr in memberExpressions.Skip(1))
                {
                    if (membObj == null)
                        return null;
                    PropertyInfo property = lastMembExpr.Type.GetProperty(membExpr.Member.Name);
                    if (property == null)
                    {
                        foreach (Type interf in lastMembExpr.Type.GetInterfaces())
                        {
                            property = interf.GetProperty(membExpr.Member.Name);
                            if (property != null)
                                break;
                        }
                    }
                    lastMembExpr = membExpr;
                    membObj = property.GetValue(membObj, null);
                }
            }

            if (methodCallExpression == null || membObj == null)
                return membObj;
            else
                return Expression.Lambda(methodCallExpression).Compile().DynamicInvoke();
        }
        //!<# } #>
    }
    //!<# if (Params.TargetFramework == "3.5") { #>

    /// <summary>
    /// Provides Enum.HasFlag implementation for .NET 3.5.
    /// </summary>
    public static class EnumExt
    {
        /// <summary>
        /// Checks if enumeration has a specific flag set.
        /// </summary>
        /// <param name="variable">Flags enumeration to check.</param>
        /// <param name="value">Flag to check for.</param>
        /// <returns>True if enumeration contains flag, false otherwise.</returns>
        public static bool HasFlag(this Enum variable, Enum value)
        {
            if (variable == null)
                return false;

            if (value == null)
                throw new ArgumentNullException("value");

            // Not as good as the .NET 4 version of this function, but still acceptable in most cases.
            if (!Enum.IsDefined(variable.GetType(), value))
            {
                throw new ArgumentException(string.Format(
                    "Enumeration type mismatch. The flag is of type '{0}', was expecting '{1}'.",
                    value.GetType(), variable.GetType()));
            }

            ulong num = Convert.ToUInt64(value);
            return ((Convert.ToUInt64(variable) & num) == num);
        }
    }
    //!<# } #>
}
