//!<#@ template language="C#" #>
//!<#@ samplevars Processor="DirectiveProcessor" #>
//!<# if (!Params.ACL)  return string.Empty; #>
//!<# if (Params.Storage != StorageType.Ntfs)  return string.Empty; #>
//!<# if ( !Params.AuthenticationType.HasFlag(AuthenticationType.Ntlm) ) return string.Empty; #>
using System.Collections.Generic;
using System.DirectoryServices.AccountManagement;
using System.Linq;
using System.Threading.Tasks;

using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Acl;

namespace ITHit.WebDav.Server.SampleModel.__NtfsStorage__.Acl
{
    /// <summary>
    /// Logical folder which contains users.
    /// It has path '/acl/users'
    /// </summary>
    public class UserFolder : LogicalFolder, IPrincipalFolderAsync
    {
        /// <summary>
        /// Path to folder which contains users.
        /// </summary>
        public static string PREFIX = AclFolder.PREFIX + "/users";

        /// <summary>
        /// Gets path of folder which contains users.
        /// </summary>
        public static string PATH
        {
            get { return PREFIX + "/"; }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="UserFolder"/> class.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/> class.</param>
        public UserFolder(DavContext context) : base(context, "users", PATH)
        {
        }

        /// <summary>
        /// Retrieves users.
        /// </summary>
        /// <param name="properties">List of properties which will be retrieved by the engine later.</param>
        /// <returns>Enumerable with users.</returns>
        public override async Task<IEnumerable<IHierarchyItemAsync>> GetChildrenAsync(IList<PropertyName> properties)
        {
            return Context.PrincipalOperation<IEnumerable<IHierarchyItemAsync>>(getUsers);
        }

        /// <summary>
        /// Retrieves all users in computer/domain.
        /// </summary>
        /// <returns>Enumerable with users.</returns>
        private IEnumerable<IHierarchyItemAsync> getUsers()
        {
            var insUserPrincipal = new UserPrincipal(Context.GetPrincipalContext());
            insUserPrincipal.Name = "*";
            var insPrincipalSearcher = new PrincipalSearcher(insUserPrincipal);

            return insPrincipalSearcher.FindAll().Select(
                u => new User((UserPrincipal)u, Context)).Cast<IHierarchyItemAsync>().ToList();
        }

        //$<IPrincipalFolderAsync.CreateFolderAsync
        /// <summary>
        /// We don't support creating folders inside this folder.
        /// </summary>        
        public async Task<IPrincipalFolderAsync> CreateFolderAsync(string name)
        {
            throw new DavException("Creating folders is not implemented", DavStatus.NOT_IMPLEMENTED);
        }
        //$>

        //$<IPrincipalFolderAsync.CreatePrincipalAsync
        /// <summary>
        /// Creates user.
        /// </summary>
        /// <param name="name">User name.</param>
        /// <returns>Newly created user.</returns>
        public async Task<IPrincipalAsync> CreatePrincipalAsync(string name)
        {
            if (!PrincipalBase.IsValidUserName(name))
            {
                throw new DavException("User name contains invalid characters", DavStatus.FORBIDDEN);
            }

            var userPrincipal = new UserPrincipal(Context.GetPrincipalContext());

            userPrincipal.Name = name;
            userPrincipal.UserPrincipalName = name;

            userPrincipal.Enabled = true;
            userPrincipal.ExpirePasswordNow();

            Context.PrincipalOperation(userPrincipal.Save);
            
            return new User(userPrincipal, Context);
        }
        //$>

        //$<IPrincipalFolderAsync.FindPrincipalsByPropertyValuesAsync
        /// <summary>
        /// Finds users whose properties have certain values.
        /// </summary>
        /// <param name="propValues">Properties and values to look for.</param>
        /// <param name="props">Properties that will be requested by the engine from the returned users.</param>
        /// <returns>Enumerable with users whose properties match.</returns>
        public async Task<IEnumerable<IPrincipalAsync>> FindPrincipalsByPropertyValuesAsync(
            IList<PropertyValue> propValues,
            IList<PropertyName> props)
        {
            var user = new UserPrincipal(Context.GetPrincipalContext());
            user.Name = "*";
            foreach (PropertyValue v in propValues)
            {
                if (v.QualifiedName == PropertyName.DISPLAYNAME)
                {
                    user.Name = "*" + v.Value + "*";
                }
            }
            
            var searcher = new PrincipalSearcher(user);
            return searcher.FindAll().Select(u => new User((UserPrincipal)u, Context)).Cast<IPrincipalAsync>();
        }
        //$>

        //$<IPrincipalFolderAsync.GetPrincipalSearcheablePropertiesAsync
        /// <summary>
        /// Returns list of properties which can be used in <see cref="FindPrincipalsByPropertyValuesAsync"/>.
        /// </summary>
        /// <returns></returns>
        public async Task<IEnumerable<PropertyDescription>> GetPrincipalSearcheablePropertiesAsync()
        {
            return new[]{ new PropertyDescription
                             {
                                 Name = PropertyName.DISPLAYNAME,                                 
                                 Description = "Principal name",
                                 Lang = "en"
                             } };
        }
        //$>

        //$<IPrincipalFolderAsync.GetMatchingPrincipalsAsync
        /// <summary>
        /// Returns <see cref="IPrincipalAsync"/> for the current user.
        /// </summary>
        /// <param name="props">Properties that will be asked later from the user returned.</param>
        /// <returns>Enumerable with users.</returns>
        public async Task<IEnumerable<IPrincipalAsync>> GetMatchingPrincipalsAsync(IList<PropertyName> props)
        {
            return new[]{ User.FromName(Context.WindowsIdentity.Name, Context) };
        }
        //$>
    }
}
