//!<#@ template language="C#" #>
//!<#@ samplevars Processor="DirectiveProcessor" #>
//!<# if (!Params.CardDav)  return string.Empty; #>
//!<# if (Params.Storage != StorageType.Ntfs)  return string.Empty; #>
using System;
using System.IO;
using System.Text.RegularExpressions;
using System.Threading.Tasks;

using ITHit.WebDAV.Server.CardDav;


namespace ITHit.WebDav.Server.SampleModel.__NtfsStorage__.CardDav
{
    /// <summary>
    /// Represents business card in an address book on a CardDAV server. Typically contains a single business card in vCard format.
    /// Instances of this class correspond to the following path: [DAVLocation]/addressbooks/[user_name]/[addressbook_name]/[file_name].vcf.
    /// </summary>
    /// <example>
    /// [DAVLocation]
    ///  |-- ...
    ///  |-- addressbooks
    ///      |-- ...
    ///      |-- [User2]
    ///           |-- [Address Book 1]
    ///           |-- ...
    ///           |-- [Address Book X]
    ///                |-- [File 1.vcf]  -- this class
    ///                |-- ...
    ///                |-- [File X.vcf]  -- this class
    /// </example>
    public class CardFile : DavFile, ICardFileAsync
    {
        /// <summary>
        /// Returns business card file that corresponds to path.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/></param>
        /// <param name="path">Encoded path relative to WebDAV root.</param>
        /// <returns>CardFile instance or null if not found.</returns>
        public static CardFile GetCardFile(DavContext context, string path)
        {
            string pattern = string.Format(@"^/?{0}/(?<user_name>[^/]+)/(?<addressbook_name>[^/]+)/(?<file_name>[^/]+\.vcf)$",
                              AddressbooksRootFolder.AddressbooksRootFolderPath.Trim(new char[] { '/' }).Replace("/", "/?"));
            if (!Regex.IsMatch(path, pattern))
                return null;

            FileInfo file = new FileInfo(context.MapPath(path));
            if (!file.Exists)
                return null;

            return new CardFile(file, context, path);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CardFile"/> class.
        /// </summary>
        /// <param name="file"><see cref="FileInfo"/> for corresponding object in file system.</param>
        /// <param name="context">Instance of <see cref="DavContext"/></param>
        /// <param name="path">Encoded path relative to WebDAV root.</param>
        private CardFile(FileInfo file, DavContext context, string path)
            : base(file, context, path)
        {
        }

        public override async Task ReadAsync(Stream output, long startIndex, long count)
        {
            await base.ReadAsync(output, startIndex, count);
        }

        public override async Task<bool> WriteAsync(Stream content, string contentType, long startIndex, long totalFileSize)
        {
            bool result = await base.WriteAsync(content, contentType, startIndex, totalFileSize);
            return result;
        }
    }
}