//!<#@ template language="C#" #>
//!<#@ samplevars Processor="DirectiveProcessor" #>
//!<# if (Params.Storage != StorageType.Ntfs)  return string.Empty; #>
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Web;
using System.Threading.Tasks;

using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Acl;
using ITHit.WebDAV.Server.Class1;
//!<# if (Params.ResumableUpload || Params.IEUpload) { #>
using ITHit.WebDAV.Server.ResumableUpload;
//!<# } #>
using ITHit.WebDav.Server.SampleModel.__NtfsStorage__.NtfsStreamsEngine;

namespace ITHit.WebDav.Server.SampleModel.__NtfsStorage__
{
    /// <summary>
    /// Represents file in WebDAV repository.
    /// </summary>
    public class DavFile : DavHierarchyItem, IFileAsync/*!<# if (Params.ResumableUpload || Params.IEUpload) { #>!*/, IResumableUploadAsync, IUploadProgressAsync /*!<# } #>!*/
    {
        /// <summary>
        /// Corresponding <see cref="FileInfo"/>.
        /// </summary>
        private readonly FileInfo fileInfo;

        /// <summary>
        /// Size of chunks to upload/download.
        /// (1Mb) buffer size used when reading and writing file content.
        /// </summary>
        private const int bufSize = 1048576;

        /// <summary>
        /// Returns file that corresponds to path.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/></param>
        /// <param name="path">Encoded path relative to WebDAV root.</param>
        /// <returns>DavFile instance or null if physical file not found in file system.</returns>
        public static DavFile GetFile(DavContext context, string path)
        {
            string filePath = context.MapPath(path);
            FileInfo file = new FileInfo(filePath);
            // to block vulnerability when "%20" folder can be injected into path and file.Exists returns 'true'
            if (!file.Exists || String.Compare(file.FullName.TrimEnd('\\'), filePath, StringComparison.OrdinalIgnoreCase) != 0)
                return null;

            return new DavFile(file, context, path);
        }

        /// <summary>
        /// Initializes a new instance of the DavFile class.
        /// </summary>
        /// <param name="file"><see cref="FileInfo"/> for corresponding object in file system.</param>
        /// <param name="context">Instance of <see cref="DavContext"/></param>
        /// <param name="path">Encoded path relative to WebDAV root.</param>
        protected DavFile(FileInfo file, DavContext context, string path)
            : base(file, context, path)
        {
            fileInfo = file;
        }

        //$<IContentAsync.ContentType
        /// <summary>
        /// Gets content type.
        /// </summary>
        public virtual string ContentType
        {
            get { return MimeType.GetMimeType(fileSystemInfo.Extension) ?? "application/octet-stream"; }
        }
        //$>

        //$<IContentAsync.ContentLength
        /// <summary>
        /// Gets length of the file.
        /// </summary>
        public long ContentLength
        {
            get { return fileInfo.Length; }
        }
        //$>

        //$<IContentAsync.Etag
        /// <summary>
        /// Gets Etag of the file.
        /// </summary>
        public string Etag
        {
            get
            {
                var serialNumber = GetStreamAndDeserialize<int>("SerialNumber", context.Logger);
                return string.Format("{0}-{1}", Modified.ToBinary(), serialNumber);
            }
        }
        //$>

        //$<IContentAsync.ReadAsync
        /// <summary>
        /// Called when a client is downloading a file. Copies file contents to ouput stream.
        /// </summary>
        /// <param name="output">Stream to copy contents to.</param>
        /// <param name="startIndex">Index from.</param>
        /// <param name="count">Number of bytes.</param>
        public virtual async Task ReadAsync(Stream output, long startIndex, long count)
        {
            //!<# if (Params.Hosting == HostingEnvironment.IIS) { #>
            //Set timeout to maximum value to be able to download large files.
            HttpContext.Current.Server.ScriptTimeout = int.MaxValue;

            //!<# } #>
            if (ContainsDownloadParam(context.Request.RawUrl))
            {
                AddContentDisposition(context, Name);
            }

            await context.FileOperationAsync(
               this,
               () => readInternalAsync(output, startIndex, count),
               Privilege.Read);
        }

        private async Task readInternalAsync(Stream output, long startIndex, long count)
        {
            var buffer = new byte[bufSize];
            using (var fileStream = fileInfo.Open(FileMode.Open, FileAccess.Read, FileShare.Read))
            {
                fileStream.Seek(startIndex, SeekOrigin.Begin);
                int bytesRead;
                var toRead = (int)Math.Min(count, bufSize);

                try
                {
                    while (toRead > 0 && (bytesRead = await fileStream.ReadAsync(buffer, 0, toRead)) > 0)
                    {
                        await output.WriteAsync(buffer, 0, bytesRead);
                        count -= bytesRead;
                    }
                }
                //!<# if (Params.Hosting == HostingEnvironment.IIS) { #>
                catch (HttpException)
                {
                    // The remote host closed the connection (for example Cancel or Pause pressed).
                }
                //!<# } else { #>
                catch (System.Net.HttpListenerException ex)
                {
                    // The remote host closed the connection (for example Cancel or Pause pressed).
                }
                //!<# } #>
            }
        }
        //$>

        //$<IContentAsync.WriteAsync
        /// <summary>
        /// Called when a file or its part is being uploaded.
        /// </summary>
        /// <param name="content">Stream with data.</param>
        /// <param name="contentType">Content type.</param>
        /// <param name="startIndex">Starting byte in target file
        /// for which data comes in <paramref name="content"/> stream.</param>
        /// <param name="totalFileSize">Size of file as it will be after all parts are uploaded. -1 if unknown (in case of chunked upload).</param>
        /// <returns>Whether the whole stream has been written. This result is used by the engine to determine
        /// if auto checkin shall be performed (if auto versioning is used).</returns>
        public virtual async Task<bool> WriteAsync(Stream content, string contentType, long startIndex, long totalFileSize)
        {
            //!<# if (Params.Class2) { #>
            await RequireHasTokenAsync();

            //!<# } #>
            //!<# if (Params.Hosting == HostingEnvironment.IIS) { #>
            //Set timeout to maximum value to be able to upload large files.
            HttpContext.Current.Server.ScriptTimeout = int.MaxValue;

            //!<# } #>
            return await context.FileOperationAsync<bool>(
                this,
                () => writeInternalAsync(content, startIndex, totalFileSize),
                Privilege.Write);
        }

        private async Task<bool> writeInternalAsync(Stream content, long startIndex, long totalLength)
        {
            if (startIndex == 0 && fileInfo.Length > 0)
            {
                SafeNativeMethods.TruncateFile(fileInfo);
            }

            //!<# if (Params.ResumableUpload || Params.IEUpload) { #>
            await RewriteStreamAsync("TotalContentLength", totalLength >= 0 ? (object)totalLength : null);
            //!<# } #>
            await RewriteStreamAsync("SerialNumber", GetStreamAndDeserialize<int>("SerialNumber", context.Logger) + 1);

            using (var fileStream = fileInfo.Open(FileMode.OpenOrCreate, FileAccess.Write, FileShare.Read))
            {
                if (fileStream.Length < startIndex)
                {
                    throw new DavException("Previous piece of file was not uploaded.", DavStatus.PRECONDITION_FAILED);
                }

                fileStream.Seek(startIndex, SeekOrigin.Begin);
                var buffer = new byte[bufSize];

                int lastBytesRead;
                try
                {
                    while ((lastBytesRead = await content.ReadAsync(buffer, 0, bufSize)) > 0)
                    {
                        await fileStream.WriteAsync(buffer, 0, lastBytesRead);
                        await fileStream.FlushAsync();
                    }
                }
                //!<# if (Params.Hosting == HostingEnvironment.IIS) { #>
                catch (HttpException)
                {
                    // The remote host closed the connection (for example Cancel or Pause pressed).
                }
                //!<# } else { #>
                catch (System.Net.HttpListenerException ex)
                {
                    // The remote host closed the connection (for example Cancel or Pause pressed).
                }
                //!<# } #>
            }

            return true;
        }
        //$>

        //$<IHierarchyItemAsync.MoveToAsync
        /// <summary>
        /// Move file to destination folder.
        /// </summary>
        /// <param name="destFolder">Folder to move to.</param>
        /// <param name="destName">New name of the file.</param>
        /// <param name="multistatus">Instance of <see cref="MultistatusException"/> where we put errors which 
        /// occurred with other items involved and continue
        /// moving other items.</param>
        public override async Task MoveToAsync(IItemCollectionAsync destFolder, string destName, MultistatusException multistatus)
        {
            //!<# if (Params.Class2) { #>
            await RequireHasTokenAsync();
            //!<# } #>
            await context.FileOperationAsync(
                () => moveToInternalAsync(destFolder, destName, multistatus));
        }

        private async Task moveToInternalAsync(IItemCollectionAsync folder, string destName, MultistatusException multistatus)
        {
            var targetFolder = (DavFolder)folder;

            if (targetFolder == null || !Directory.Exists(targetFolder.GetFullPath()))
            {
                throw new DavException("Target directory doesn't exist", DavStatus.CONFLICT);
            }

            var newDirPath = System.IO.Path.Combine(targetFolder.GetFullPath(), destName);
            var targetPath = targetFolder.Path + EncodeUtil.EncodeUrlPart(destName);

            //If an item with the same name exists in target directory - remove it.
            try
            {
                IHierarchyItemAsync item = await context.GetHierarchyItemAsync(targetPath);
                if (item != null)
                    await item.DeleteAsync(multistatus);
            }
            catch (DavException ex)
            {
                //Report exception to client and continue with other items by returning from recursion.
                multistatus.AddInnerException(targetPath, ex);
                return;
            }

            //Move the file.
            try
            {
                File.Move(fileSystemInfo.FullName, newDirPath);
            }
            catch (UnauthorizedAccessException)
            {
                //Exception occurred with the item for which MoveTo was called - fail the operation.
                var ex = new NeedPrivilegesException("Not enough privileges");
                ex.AddRequiredPrivilege(targetPath, Privilege.Bind);

                string parentPath = System.IO.Path.GetDirectoryName(Path);
                ex.AddRequiredPrivilege(parentPath, Privilege.Unbind);
                throw ex;
            }
        }
        //$>

        //$<IHierarchyItemAsync.DeleteAsync
        /// <summary>
        /// Delete the file.
        /// </summary>
        /// <param name="multistatus">We don't use it here.</param>
        public override async Task DeleteAsync(MultistatusException multistatus)
        {
            //!<# if (Params.Class2) { #>
            await RequireHasTokenAsync();

            //!<# } #>
            context.FileOperation(
                this,
                () => fileSystemInfo.Delete(),
                Privilege.Unbind);
        }
        //$>

        //$<IHierarchyItemAsync.CopyToAsync
        /// <summary>
        /// Copies file to destination folder.
        /// </summary>
        /// <param name="destFolder">Destination folder.</param>
        /// <param name="destName">New file name.</param>
        /// <param name="deep">We don't use it here.</param>
        /// <param name="multistatus">We report exceptions which occurred with other items here.</param>
        public override async Task CopyToAsync(IItemCollectionAsync destFolder, string destName, bool deep, MultistatusException multistatus)
        {
            await context.FileOperationAsync(
                () => copyToInternalAsync(destFolder, destName, multistatus));
        }

        private async Task copyToInternalAsync(IItemCollectionAsync folder, string destName, MultistatusException multistatus)
        {
            var targetFolder = (DavFolder)folder;

            if (targetFolder == null || !Directory.Exists(targetFolder.GetFullPath()))
            {
                throw new DavException("Target directory doesn't exist", DavStatus.CONFLICT);
            }

            var newDirPath = System.IO.Path.Combine(targetFolder.GetFullPath(), destName);
            var targetPath = targetFolder.Path + EncodeUtil.EncodeUrlPart(destName);

            //If an item with the same name exists - remove it.
            try
            {
                IHierarchyItemAsync item = await context.GetHierarchyItemAsync(targetPath);
                if (item != null)
                    await item.DeleteAsync(multistatus);
            }
            catch (DavException ex)
            {
                //Report error with other item to client.
                multistatus.AddInnerException(targetPath, ex);
                return;
            }

            //Copy the file.
            try
            {

                //!<# if (Params.TargetFramework45OrLater) { #>
                using (FileStream sourceStream = File.Open(fileSystemInfo.FullName, FileMode.Open))
                {
                    using (FileStream destinationStream = File.Create(newDirPath))
                    {
                        await sourceStream.CopyToAsync(destinationStream);
                    }
                }
                //!<# }else{ #>
                File.Copy(fileSystemInfo.FullName, newDirPath);
                //!<# } #>

                DeleteStream(newDirPath, "Locks");
            }
            catch (UnauthorizedAccessException)
            {
                //Fail
                var ex = new NeedPrivilegesException("Not enough privileges");
                string parentPath = System.IO.Path.GetDirectoryName(Path);
                ex.AddRequiredPrivilege(parentPath, Privilege.Bind);
                throw ex;
            }
        }
        //$>

        //!<# if (Params.ResumableUpload || Params.IEUpload) { #>

        //$<IResumableUploadAsync.CancelUploadAsync
        /// <summary>
        /// Cancels upload.
        /// </summary>
        /// <remarks>
        /// Client do not plan to restore upload. Remove any temporary files / cleanup resources here.
        /// </remarks>
        public async Task CancelUploadAsync()
        {
            //!<# if (!Params.DeltaV) { #>
            await DeleteAsync(null);
            //!<# } else if (false) { #>
            // TODO: Uncheckout()
            //!<# } #>
        }
        //$>

        //$<IResumableUploadAsync.LastChunkSaved
        /// <summary>
        /// Gets date when last chunk was saved to this file.
        /// </summary>
        public DateTime LastChunkSaved
        {
            get { return fileInfo.Exists ? fileInfo.LastWriteTimeUtc : DateTime.MinValue; }
        }
        //$>

        //$<IResumableUploadAsync.BytesUploaded
        /// <summary>
        /// Gets number of bytes uploaded sofar.
        /// </summary>
        public long BytesUploaded
        {
            get { return ContentLength; }
        }
        //$>

        //$<IResumableUploadAsync.TotalContentLength
        /// <summary>
        /// Gets total length of the file being uploaded.
        /// </summary>
        public long TotalContentLength
        {
            get { return GetStreamAndDeserialize<long>("TotalContentLength", context.Logger); }
        }
        //$>

        //$<IUploadProgressAsync.GetUploadProgressAsync
        /// <summary>
        /// Return instance of <see cref="IUploadProgressAsync"/> interface.
        /// </summary>
        /// <returns>Just returns this class.</returns>
        public async Task<IEnumerable<IResumableUploadAsync>> GetUploadProgressAsync()
        {
            return new[] { this };
        }
        //$>

        //!<# } #>

        /// <summary>
        /// Serializes object <paramref name="o"/> and saves it into specified NTFS stream.
        /// </summary>
        /// <remarks>
        /// Preserves LastModified time.
        /// </remarks>
        /// <typeparam name="T">Type of object to be serialized.</typeparam>
        /// <param name="streamName">NTFS stream name.</param>
        /// <param name="o">Object to be serialized.</param>
        protected override async Task RewriteStreamAsync<T>(string streamName, T o)
        {
            await context.FileOperationAsync(
               this,
               () => rewriteStreamAsync(streamName, o),
               Privilege.Write);
        }

        /// <summary>
        /// Deletes NTFS stream with specified name.
        /// </summary>
        /// <remarks>
        /// Preserves LastModified time.
        /// </remarks>
        /// <param name="fileName">File name.</param>
        /// <param name="streamName">NTFS stream name.</param>
        protected override void DeleteStream(string fileName, string streamName)
        {
            DateTime lastWriteTimeUtc = File.GetLastWriteTimeUtc(fileName);


            new AlternateDataStreamInfo(fileName, streamName).Delete();

            // Preserve last modification date.
            SafeNativeMethods.SetLastWriteTimeUtc(fileName, lastWriteTimeUtc);
        }

        internal static bool ContainsDownloadParam(string url)
        {
            int ind = url.IndexOf('?');
            if (ind > 0 && ind < url.Length - 1)
            {
                var param = url.Substring(ind + 1).Split('&');
                return param.Any(p => p.StartsWith("download"));
            }
            return false;
        }

        internal static void AddContentDisposition(DavContext context, string name)
        {
            if (context.Request.UserAgent.Contains("MSIE")) // problem with file extensions in IE
            {
                var fileName = EncodeUtil.EncodeUrlPart(name);
                context.Response.AddHeader("Content-Disposition", "attachment; filename=\"" + fileName + "\"");
            }
            else
                context.Response.AddHeader("Content-Disposition", "attachment;");
        }

        private async Task rewriteStreamAsync<T>(string streamName, T o)
        {
            DateTime lastWriteTimeUtc = fileSystemInfo.LastWriteTimeUtc;

            await SerializationUtils.RewriteStreamAsync(
                       streamName,
                       fileSystemInfo,
                       o != null ? SerializationUtils.Serialize(o) : null);

            // Preserve last modification date.
            SafeNativeMethods.SetLastWriteTimeUtc(fileSystemInfo.FullName, lastWriteTimeUtc);
        }
    }
}
